"""
Base parameter class
"""
import yaml
from astropy import units as u
from pypsg.cfg.base import Table

class BaseParameters:
    """
    Base class for Parameters
    """
    def __init__(self,**kwargs):
        for key,value in kwargs.items():
            self.__setattr__(key,value)
    @classmethod
    def _from_dict(cls,d:dict):
        return cls(**d)
    @classmethod
    def from_dict(cls,d:dict,*args):
        """
        Construct a BaseParameters (or subclass) instance from a dictionary.

        Parameters
        ----------
        d : dict
            The dictionary containing the parameters for `LimbDarkeningParameters`.

        Returns
        -------
        BaseParameters or subclass
            An instance of `BaseParameters` initialized with the provided parameters.

        Notes
        -----
        If the dictionary contains a key named 'preset', the corresponding class method
        will be called to create an instance with preset parameters.
        If 'preset' key is not present, the dictionary is expected to contain the values
        needed by the constructor.

        """
        if 'preset' in d.keys():
            return getattr(cls,d['preset'].replace('-','_'))()
        else:
            return cls._from_dict(d,*args)
    @classmethod
    def from_preset(cls,name):
        """
        Load a ``BaseParameters`` instance from a preset file.

        Parameters
        ----------
        name : str
            The name of the preset to load.
        
        Returns
        -------
        BaseParameters
            The class instance loaded from a preset.
        """
        if hasattr(cls,'_PRESET_PATH'):
            with open(cls._PRESET_PATH, 'r',encoding='UTF-8') as file:
                data = yaml.safe_load(file)
                return cls.from_dict(data[name])
        else:
            raise NotImplementedError('This class does not have a ``_PRESET_PATH`` attribute.')


class PSGtable(BaseParameters):
    """
    Class to store Table data for PSG
    """
    def __init__(
        self,
        x:u.Quantity,
        y:u.Quantity
    ):
        self.x = x
        self.y = y
    def __str__(self):
        pairs = []
        for x,y in zip(self.x,self.y):
            pairs.append(f'{y:.2e}@{x:.2e}')
        return ','.join(pairs)
    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            x = d['x']*u.Unit(d.get('xunit','')),
            y = d['y']*u.Unit(d.get('yunit',''))
        )
    @classmethod
    def from_dict(cls, d: dict, *args):
        """
        Construct a ``PSGtable`` object from a dictionary.
        
        Parameters
        ----------
        d : dict
            The dictionary to use to construct the class.

        Returns
        -------
        PSGtable
            The constructed class instance.

        """
        return super().from_dict(d, *args)
    def to_psg(self):
        """
        Convert to pypsg table
        """
        return Table(
            x = self.x.to_value(u.dimensionless_unscaled) if self.x.unit is u.dimensionless_unscaled else self.x,
            y = self.y.to_value(u.dimensionless_unscaled) if self.y.unit is u.dimensionless_unscaled else self.y
        )

def parse_table(val,cls):
    """
    Parse some input that could potentially construct a ``PSGtable`` object.
    
    If ``val`` is a dictionary, use it to construct a table. Otherwise, construct
    an object of type ``cls`` from ``val`` (which is probably a string).

    Parameters
    ----------
    val : any
        The input to parse.
    cls : type
        The class to cast ``val`` to if it is not a dictionary.

    Returns
    -------
    any
        The parsed input.
    """
    if isinstance(val,dict):
        return PSGtable.from_dict(val['table']).to_psg()
    else:
        return cls(val)



"""
VSPEC configurations

This module contains global configurations used in the VSPEC code.
"""
from astropy import units as u
import numpy as np
from pathlib import Path

MSH = u.def_unit('msh', 1e-6 * 0.5 * 4*np.pi*u.R_sun**2)
"""
Micro-solar hemisphere

This is a standard unit in heliophysics that
equals one millionth of one half the surface area of the Sun.

:type: astropy.units.Unit
"""

stellar_area_unit = MSH
"""
The standard stellar surface area unit.

This unit is used to represent the surface area of stars in VSPEC.
The micro-solar hemisphere is chosen because most Sun Spot literature uses
this unit.

:type: astropy.units.Unit
"""

starspot_initial_area = 10*MSH
"""
Initial ``StarSpot`` area.

Because spots grow exponentially, they can't start at 0 area.
When they are born they are given this small area.

:type: astropy.units.Quantity

.. todo::
    This should optionaly be set by the user. So that smaller
    star spot area regimes are accessible.
"""

flux_unit = u.Unit('W m-2 um-1')
"""
The standard unit of flux.

This unit is used to standardize the flux in VSPEC calculations.
:math:`W m^{-2} \\mu m^{-1}` is chosen because it is the standard
spectral irrandience unit in PSG.

:type: astropy.units.Unit
"""

teff_unit = u.K
"""
The standard unit of temperature.

This selection standardizes units across the package.

:type: astropy.units.Unit
"""

wl_unit = u.um
"""
The standard unit of wavelength.

The output wavelength can still be changed by the user, but internally
we want units to be consistent.

:type: astropy.units.Unit
"""

nlat = 500
"""
The default latitude resolution for the stellar model. This should
be set by finding a balance between noticing small changes in spots/faculae
and computation time.

:type: int
"""

nlon = 1000
"""
The default longitude resolution for the stellar model. This should
be set by finding a balance between noticing small changes in spots/faculae
and computation time.

:type: int
"""

grid_teff_bounds = (2300*u.K, 3900*u.K)
"""
The limits on the effective temperature allowed by the grid.

:type: tuple of astropy.units.Quantity
"""

planet_distance_unit = u.AU
"""
The standard unit of planetary semimajor axis.

This unit is determined by PSG and used to standardize
the semimajor axis of planets in VSPEC.

:type: astropy.units.Unit
"""

planet_radius_unit = u.km
"""
The standard unit of planet radius.

This unit is determined by PSG and used to standardize
the radius of planets in VSPEC.

:type: astropy.units.Unit
"""

period_unit = u.day
"""
The standard unit of planet orbital period.

This unit is determined by PSG and used to standardize
the orbital and rotational periods of planets in VSPEC.

:type: astropy.units.Unit
"""

psg_encoding = 'UTF-8'
"""
Default encoding for files from PSG.

:type: str
"""

PSG_PORT = 3000
"""
Default port to run PSG locally.

This port number is used to access the PSG API locally.

:type: int
"""

PSG_EXT_URL = 'https://psg.gsfc.nasa.gov'
"""
The URL of the external PSG (Planetary Spectrum Generator) version.

This URL is used to access the PSG web interface externally.

.. warning::
    The external PSG server only allows 100 anonymous API
    calls per day. Please either use an API key or a local
    PSG installation.

:type: str
"""

PSG_CFG_MAX_LINES = 1500
"""
The maximum number of lines to allow in
the PSG config file. Do not set this over 2000,
as PSG will stop updating.

:type: int

Notes
-----
For security reasons, PSG only allows ``.cfg`` files to be
up to 2000 lines long. This means that, every so often,
we have to send a ``set`` command, which resets everything.
This command is sent after the ``.cfg`` returned by PSG reaches
`PSG_CFG_MAX_LINES` lines long.
"""


psg_pressure_unit = u.bar
"""
PSG atmospheric pressure unit.

This unit is determined by PSG and used to standardize
the atmospheric pressure of planets in VSPEC.

:type: astropy.units.Unit
"""
psg_aerosol_size_unit = u.m
"""
PSG aerosol size unit.

This unit is determined by PSG and used to
standardize aerosol size in VSPEC.

:type: astropy.units.Unit
"""

atmosphere_type_dict = {'H2':45,'He':0,'H2O':1,'CO2':2,'O3':3,'N2O':4,'CO':5,'CH4':6,'O2':7,
                        'NO':8,'SO2':9,'NO2':10,'N2':22,'HNO3':12,'HO2NO2':'SEC[26404-66-0] Peroxynitric acid',
                        'N2O5':'XSEC[10102-03-1] Dinitrogen pentoxide','O':'KZ[08] Oxygen',
                        'OH':'EXO[OH]'}
"""
A dictionary mapping molecular species to the default
database to use to create opacities. These are all
internal to PSG, but must be set by ``VSPEC``.

Integers mean that we want to use data from the HITRAN database, which for a number ``N``
is represented in PSG by ``HIT[N]``. Strings are sent straight to PSG as is.

:type: dict
"""

aerosol_name_dict = {
    'Water':{
        'name':'CLDLIQ',
        'size':'REL'
    },
    'WaterIce':{
        'name':'CLDICE',
        'size':'REI'
    }
}
"""
A dictionary mapping aerosol species from their PSG name
to their name in the WACCM NetCDF format.

:type: dict
"""

aerosol_type_dict = {
    'Water': 'AFCRL_Water_HRI',
    'WaterIce': 'Warren_ice_HRI'
}
"""
A dictionary mapping aerosol species to the default
database to use. These are all
internal to PSG, but must be set by ``VSPEC``.

:type: dict
"""
############################
# Paths and file conventions
############################

N_ZFILL = 5
"""
``__width`` argument for filename ``str.zfill()`` calls. When writing
and reading files in the `VSPEC` output, this variable specifies
the number of leading zeros to use in the filename.

:type: int
"""

RAW_PHOENIX_PATH = Path(__file__).parent / 'data' / 'NextGenModels' / 'RawData'
"""
The path to the raw PHOENIX stellar models.

:type: pathlib.Path
"""

BINNED_PHOENIX_PATH = Path(__file__).parent / 'data' / 'NextGenModels' / 'binned'
"""
The path to the binned PHOENIX stellar models.

:type: pathlib.Path
"""

EXAMPLE_GCM_PATH = Path(__file__).parent / 'data' / 'GCMs'
"""
The path to example GCMs.

:type: pathlib.Path
"""

MOLEC_DATA_PATH = Path(__file__).parent / 'data' / 'molec.json'
"""
The path to the file containing molecular data for analysis.

:type: pathlib.Path
"""
PRESET_PATH = Path(__file__).parent / 'presets'
"""
The path to parameter presets.

:type: pathlib.Path
"""

VSPEC_PARENT_PATH = Path('.vspec')


"""
Module to read parameters
"""
from pathlib import Path
import yaml
from astropy import units as u
from pypsg import PyConfig
from pypsg.cfg import models
from pypsg.globes import PyGCM
from pypsg.units import resolving_power as u_rp

from VSPEC import config
from VSPEC.params.base import BaseParameters
from VSPEC.params.stellar import StarParameters
from VSPEC.params.planet import PlanetParameters, SystemParameters
from VSPEC.params.gcm import gcmParameters, psgParameters
from VSPEC.params.observation import InstrumentParameters, ObservationParameters, SingleDishParameters, CoronagraphParameters


class Header(BaseParameters):
    """
    Header for VSPEC simulation

    Parameters
    ----------
    data_path : pathlib.Path
        The path to store run data.
    max_teff : astropy.units.Quantity
        The maximum Teff to bin.
    min_teff : astropy.units.Quantity
        The minimum Teff to bin.
    seed : int, default=None
        The seed for the random number generator.
    verbose : int, default=1,
        The level of verbosity for the simulation.
    desc : str, default=None
        A description of the run.

    Attributes
    ----------
    data_path : pathlib.Path
        The path to store run data.
    max_teff : astropy.units.Quantity
        The maximum Teff to bin.
    min_teff : astropy.units.Quantity
        The minimum Teff to bin.
    seed : int or None
        The seed for the random number generator.
    verbose : int
        The level of verbosity for the simulation.
    desc : str or None
        A description of the run.

    """

    def __init__(
        self,
        data_path: Path,
        teff_min: u.Quantity,
        teff_max: u.Quantity,
        seed: int,
        verbose: int = 1,
        desc: str = None
    ):
        self.data_path = data_path
        self.teff_min = teff_min
        self.teff_max = teff_max
        self.seed = seed
        self.verbose = verbose
        self.desc = desc

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            data_path=config.VSPEC_PARENT_PATH / d['data_path'],
            teff_min=u.Quantity(d['teff_min']),
            teff_max=u.Quantity(d['teff_max']),
            seed=None if d.get('seed', None) is None else int(
                d.get('seed', None)),
            desc=None if d.get('desc', None) is None else str(
                d.get('desc', None))
        )


class InternalParameters(BaseParameters):
    """
    Class to store parameters for a VSPEC simulation.

    Parameters
    ----------
    header : Header
        The VSPEC simulation header.
    star : StarParameters
        The parameters related to the star.
    planet : PlanetParameters
        The parameters related to the planet.
    system : SystemParameters
        The parameters related to the system.
    obs : ObservationParameters
        The parameters related to the observation.
    psg : psgParameters
        The PSG parameters.
    inst : InstrumentParameters
        The instrument parameters.
    gcm : gcmParameters
        The GCM parameters.

    Attributes
    ----------
    header : Header
        The VSPEC simulation header.
    star : StarParameters
        The parameters related to the star.
    planet : PlanetParameters
        The parameters related to the planet.
    system : SystemParameters
        The parameters related to the system.
    obs : ObservationParameters
        The parameters related to the observation.
    psg : psgParameters
        The PSG parameters.
    inst : InstrumentParameters
        The instrument parameters.
    gcm : gcmParameters
        The GCM parameters.


    """

    def __init__(
        self,
        header: Header,
        star: StarParameters,
        planet: PlanetParameters,
        system: SystemParameters,
        obs: ObservationParameters,
        psg: psgParameters,
        inst: InstrumentParameters,
        gcm: gcmParameters
    ):
        self.header = header
        self.star = star
        self.planet = planet
        self.system = system
        self.obs = obs
        self.psg = psg
        self.inst = inst
        self.gcm = gcm

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            header=Header.from_dict(d['header']),
            star=StarParameters.from_dict(d['star']),
            planet=PlanetParameters.from_dict(d['planet']),
            system=SystemParameters.from_dict(d['system']),
            obs=ObservationParameters.from_dict(d['obs']),
            psg=psgParameters.from_dict(d['psg']),
            inst=InstrumentParameters.from_dict(d['inst']),
            gcm=gcmParameters.from_dict(d)
        )

    @classmethod
    def from_dict(cls, d: dict) -> 'InternalParameters':
        """
        Create and `InternalParameters` instance from a dictionary.

        Parameters
        ----------
        d : dict
            The dictionary to construct the class from.

        Returns
        -------
        InternalParameters
            An instance of `InternalParameters`.
        """
        return super().from_dict(d)

    @classmethod
    def from_yaml(cls, path: Path) -> 'InternalParameters':
        """
        Create an `InternalParameters` instance from a YAML file.

        Parameters
        ----------
        path : Path
            The path to the YAML file.

        Returns
        -------
        InternalParameters
            An instance of the `InternalParameters` class.

        """
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
        return cls.from_dict(data)

    @property
    def target(self):
        """
        Get the target model.
        """
        target = models.Target(
            object='Exoplanet',
            name=self.planet.name,
            date=None,
            diameter=2*self.planet.radius,
            gravity=self.planet.gravity.value,
            star_distance=self.planet.semimajor_axis,
            star_velocity=None,
            solar_longitude=None,
            solar_latitude=None,
            season=None,
            inclination=self.system.inclination,
            position_angle=None,
            star_type=self.star.psg_star_template,
            star_temperature=self.star.teff,
            star_radius=self.star.radius,
            star_metallicity=None,
            obs_longitude=None,
            obs_latitude=None,
            obs_velocity=None,
            period=self.planet.orbit_period,
            orbit=None
        )
        return target

    @property
    def geometry(self):
        """
        Get the geometry model.
        """
        return models.Observatory(
            ref=None,
            offset=None,
            observer_altitude=self.system.distance,
            azimuth=None,
            stellar_type=self.star.psg_star_template,
            stellar_temperature=self.star.teff,
            stellar_magnitude=None,
            disk_angles=None,
        )

    @property
    def atmosphere(self):
        """
        Get the atmosphere model.
        """
        return models.EquilibriumAtmosphere(
            weight=self.gcm.mean_molec_weight*u.Unit('g/mol'),
            continuum=','.join(self.psg.continuum),
            nmax=self.psg.nmax,
            lmax=self.psg.lmax
        )

    @property
    def surface(self):
        """
        Get the surface model.
        """
        return models.Surface(
            albedo=None,
            temperature=None,
            emissivity=None
        )

    @property
    def generator(self):
        """
        Get the generator model.
        """
        return models.Generator(
            resolution_kernel=None,
            gas_model=self.psg.use_molecular_signatures,
            continuum_stellar=self.psg.use_continuum_stellar,
            apply_telluric_noise=None,
            apply_telluric_obs=None,
            telluric_params=None,
            rad_units=config.flux_unit,
            log_rad=None,
            gcm_binning=self.psg.gcm_binning
        )

    @property
    def telescope(self):
        """
        Get the telescope model.
        """
        if isinstance(self.inst.telescope, SingleDishParameters):
            return models.SingleTelescope(
                apperture=self.inst.telescope.aperture,
                zodi=self.inst.telescope.zodi,
                fov=self.inst.detector.beam_width,
                range1=self.inst.bandpass.wl_blue,
                range2=self.inst.bandpass.wl_red,
                resolution=self.inst.bandpass.resolving_power*u_rp
            )
        elif isinstance(self.inst.telescope, CoronagraphParameters):
            return models.Coronagraph(
                apperture=self.inst.telescope.aperture,
                zodi=self.inst.telescope.zodi,
                fov=self.inst.detector.beam_width,
                range1=self.inst.bandpass.wl_blue,
                range2=self.inst.bandpass.wl_red,
                resolution=self.inst.bandpass.resolving_power*u_rp,
                contrast=self.inst.telescope.contrast,
                iwa=self.inst.telescope.iwa
            )
        else:
            raise ValueError('Unknown telescope type')

    @property
    def noise(self):
        """
        Get the noise model.
        """
        return models.CCD(
            exp_time=self.inst.detector.integration_time,
            n_frames=int(round(
                (self.obs.integration_time/self.inst.detector.integration_time).to_value(u.dimensionless_unscaled))),
            n_pixels=self.inst.detector.ccd.pixel_sampling,
            read_noise=self.inst.detector.ccd.read_noise,
            dark_current=self.inst.detector.ccd.dark_current,
            thoughput=self.inst.detector.ccd.throughput,
            emissivity=self.inst.detector.ccd.emissivity,
            temperature=self.inst.detector.ccd.temperature,
            pixel_depth=None
        )

    def to_pyconfig(self):
        """
        Write VSPEC parameters into a `pypsg` `PyConfig` object.
        """
        return PyConfig(
            target=self.target,
            geometry=self.geometry,
            atmosphere=self.atmosphere,
            surface=self.surface,
            generator=self.generator,
            telescope=self.telescope,
            noise=self.noise,
            gcm=None
        )

    @property
    def flux_correction(self) -> float:
        """
        The flux correction for the stellar radius and distance.

        Returns
        -------
        float
            Correction for the solid angle of the star.
        """
        return (self.star.radius/self.system.distance).to_value(u.dimensionless_unscaled)**2

    @property
    def star_total_images(self) -> int:
        """
        The number of epochs to simulate for the stellar model.

        Returns
        -------
        int
            The total number of epochs to simulate the star.
        """
        return self.obs.total_images

    @property
    def planet_total_images(self) -> int:
        """
        The number of epochs to simulate for the planet model.

        Returns
        -------
        int
            The total number of epochs to simulate the planet.
        """
        return self.obs.total_images // self.psg.phase_binning



"""
GCM parameters module
"""
from typing import Callable
from pathlib import Path
from astropy import units as u
from netCDF4 import Dataset

from pypsg import PyConfig
from pypsg.globes import waccm_to_pygcm, PyGCM, exocam_to_pygcm, GCMdecoder, exoplasim_to_pygcm
from pypsg.globes.exocam import exocam
from pypsg.globes.waccm import waccm

from VSPEC.params.base import BaseParameters
from VSPEC.gcm.heat_transfer import to_pygcm as vspec_to_pygcm


def parse_molec_list(molec_list: list):
    """
    Annoyingly YAML automatically converts NO (nitrous oxide) to the boolean False.
    """
    return [mol if mol is not False else 'NO' for mol in molec_list]


class gcmParameters(BaseParameters):
    """
    Class to store GCM parameters.

    Parameters
    ----------
    gcm_getter : Callable
        A function that returns a PyGCM instance.
    mean_molec_weight : float
        The mean molecular weight of the atmosphere in g/mol.
    is_staic : bool
        If true, the GCM does not change over time. In the case that this is false a time
        parameter will be passed to the `gcm_getter`.

    """
    _defaults = {
        'lat_redistribution': 0.0
    }

    def __init__(
        self,
        gcm_getter: Callable[..., PyGCM],
        mean_molec_weight: float,
        is_static: bool
    ):
        self.get_gcm = gcm_getter
        self.mean_molec_weight = mean_molec_weight
        self.is_staic = is_static

    def content(self, **kwargs):
        """
        Get a bytes representation of the GCM.
        """
        return self.get_gcm(**kwargs).content

    def to_pycfg(self, **kwargs) -> PyConfig:
        """
        Get `pypsg.PyConfig` representation of the GCM.

        Parameters
        ----------
        obs_time : astropy.time.Time, optional
            The time of the observation. Necessary for a waccm GCM.
        """
        _gcm = self.get_gcm(**kwargs)
        return PyConfig(
            atmosphere=_gcm.update_params(),
            gcm=_gcm
        )

    @classmethod
    def _from_dict(cls, d: dict):
        gcm_dict = d['gcm']
        star_dict = d['star']
        planet_dict = d['planet']
        mean_molec_weight = float(gcm_dict['mean_molec_weight'])
        if 'binary' in gcm_dict:
            args_dict: dict = gcm_dict['binary']
            path = Path(args_dict['path']).expanduser()

            def fun():
                return PyGCM.from_decoder(GCMdecoder.from_psg(path))
            return cls(
                gcm_getter=fun,
                mean_molec_weight=mean_molec_weight,
                is_static=True
            )
        elif 'vspec' in gcm_dict:
            args_dict: dict = gcm_dict['vspec']

            def fun():
                return vspec_to_pygcm(
                    shape=(
                        int(args_dict['nlayer']),
                        int(args_dict['nlon']),
                        int(args_dict['nlat'])
                    ),
                    epsilon=float(args_dict['epsilon']),
                    star_teff=u.Quantity(star_dict['teff']),
                    r_star=u.Quantity(star_dict['radius']),
                    r_orbit=u.Quantity(planet_dict['semimajor_axis']),
                    lat_redistribution=float(args_dict.get('lat_redistribution', cls._defaults['lat_redistribution'])),
                    p_surf=u.Quantity(args_dict['psurf']),
                    p_stop=u.Quantity(args_dict['ptop']),
                    wind_u=u.Quantity(args_dict['wind']['U']),
                    wind_v=u.Quantity(args_dict['wind']['V']),
                    gamma=float(args_dict['gamma']),
                    albedo=u.Quantity(args_dict['albedo']),
                    emissivity=u.Quantity(args_dict['emissivity']),
                    molecules=args_dict['molecules'],
                )
            return cls(
                gcm_getter=fun,
                mean_molec_weight=mean_molec_weight,
                is_static=True
            )
        elif 'waccm' in gcm_dict:
            args_dict: dict = gcm_dict['waccm']
            path = Path(args_dict['path']).expanduser()
            itime = int(args_dict['itime']) if 'itime' in args_dict else None
            is_static = itime is not None
            if is_static:
                def fun():
                    with Dataset(path) as data:
                        return waccm_to_pygcm(
                            data=data,
                            itime=itime,
                            molecules=parse_molec_list(args_dict['molecules']),
                            aerosols=args_dict['aerosols'],
                            background=args_dict.get('background', None),
                            lon_start=args_dict.get('lon_start', -180),
                            lat_start=args_dict.get('lat_start', -90)
                        )
            else:
                def fun(obs_time: u.Quantity):
                    with Dataset(path) as data:
                        return waccm_to_pygcm(
                            data=data,
                            itime=waccm.get_time_index(
                                data, obs_time + u.Quantity(args_dict['tstart'])),
                            molecules=parse_molec_list(args_dict['molecules']),
                            aerosols=args_dict.get('aerosols',None),
                            background=args_dict.get('background', None),
                            lon_start=args_dict.get('lon_start', -180),
                            lat_start=args_dict.get('lat_start', -90)
                        )
            return cls(
                gcm_getter=fun,
                mean_molec_weight=mean_molec_weight,
                is_static=is_static
            )
        elif 'exocam' in gcm_dict:
            args_dict: dict = gcm_dict['exocam']
            path = Path(args_dict['path']).expanduser()
            itime = int(args_dict['itime']) if 'itime' in args_dict else None
            is_static = itime is not None
            if is_static:
                def fun():
                    with Dataset(path) as data:
                        return exocam_to_pygcm(
                            data=data,
                            itime=itime,
                            molecules=parse_molec_list(args_dict['molecules']),
                            aerosols=args_dict['aerosols'],
                            background=args_dict.get('background', None),
                            lon_start=args_dict.get('lon_start', -180),
                            lat_start=args_dict.get('lat_start', -90),
                            mean_molecular_mass=args_dict.get('mean_molecular_mass',None)
                        )
            else:
                def fun(obs_time: u.Quantity):
                    with Dataset(path) as data:
                        return exocam_to_pygcm(
                            data=data,
                            itime=exocam.get_time_index(data, obs_time),
                            molecules=parse_molec_list(args_dict['molecules']),
                            aerosols=args_dict['aerosols'],
                            background=args_dict.get('background', None),
                            lon_start=args_dict.get('lon_start', -180),
                            lat_start=args_dict.get('lat_start', -90),
                            mean_molecular_mass=args_dict.get('mean_molecular_mass',None)
                        )
            return cls(
                gcm_getter=fun,
                mean_molec_weight=mean_molec_weight,
                is_static=is_static
            )
        elif 'exoplasim' in gcm_dict:
            args_dict: dict = gcm_dict['exoplasim']
            path = Path(args_dict['path']).expanduser()

            def fun():
                with Dataset(path) as data:
                    return exoplasim_to_pygcm(
                        data=data,
                        itime=int(args_dict['itime']),
                        molecules=parse_molec_list(args_dict['molecules']),
                        aerosols=args_dict['aerosols'],
                        background=args_dict.get('background', None),
                        lon_start=args_dict.get('lon_start', -180),
                        lat_start=args_dict.get('lat_start', -90),
                        mean_molecular_mass=args_dict.get('mean_molecular_mass',None)
                    )
            return cls(
                gcm_getter=fun,
                mean_molec_weight=mean_molec_weight,
                is_static=True
            )

        else:
            raise KeyError(
                f'`binary`, `waccm`, or `vspec` not in {list(d.keys())}')


class psgParameters(BaseParameters):
    """
    Class to store parameters for the Planetary Spectrum Generator (PSG).

    Parameters
    ----------
    gcm_binning : int
        Number of spatial points to bin together in the GCM data. Use 3 for science.
    phase_binning : int
        Number of phase epochs to bin together when simulating the planet. These are later
        interpolated to match the cadence of the variable star simulation.
    use_molecular_signatures : bool
        Whether to use molecular signatures (PSG atmosphere) or not.
    use_continuum_stellar : bool
        Whether to include the stellar contiuum or not.
    nmax : int
        PSG handbook: 'When performing scattering aerosols calculations, this
        parameter indicates the number of n-stream pairs - Use 0 for extinction
        calculations only (e.g. transit, occultation)' :cite:p:`2022fpsg.book.....V`
    lmax : int
        PSG handbook: 'When performing scattering aerosols calculations, this
        parameter indicates the number of scattering Legendre polynomials used
        for describing the phase function - Use 0 for extinction calculations
        only (e.g. transit, occultation)' :cite:p:`2022fpsg.book.....V`
    continuum : list of str
        The continuum opacities to include in the radiative transfer calculation, such as
        'Rayleigh', 'Refraction', 'CIA_all'.

    Attributes
    ----------
    gcm_binning : int
        Number of spatial points to bin together in the GCM data.
    phase_binning : int
        Number of phase epochs to bin together when simulating the planet.
    use_molecular_signatures : bool
        Whether to use molecular signatures (PSG atmosphere) or not.
    use_continuum_stellar : bool
        Whether to include the stellar contiuum or not.
    nmax : int
        PSG handbook: 'When performing scattering aerosols calculations, this
        parameter indicates the number of n-stream pairs - Use 0 for extinction
        calculations only (e.g. transit, occultation)' :cite:p:`2022fpsg.book.....V`
    lmax : int
        PSG handbook: 'When performing scattering aerosols calculations, this
        parameter indicates the number of scattering Legendre polynomials used
        for describing the phase function - Use 0 for extinction calculations
        only (e.g. transit, occultation)' :cite:p:`2022fpsg.book.....V`

    """
    _defaults = {
        'use_stellar_continuum': True
    }

    def __init__(
        self,
        gcm_binning: int,
        phase_binning: int,
        use_molecular_signatures: bool,
        use_continuum_stellar: bool,
        nmax: int,
        lmax: int,
        continuum: list
    ):
        self.gcm_binning = gcm_binning
        self.phase_binning = phase_binning
        self.use_molecular_signatures = use_molecular_signatures
        self.use_continuum_stellar = use_continuum_stellar
        self.nmax = nmax
        self.lmax = lmax
        self.continuum = continuum

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            gcm_binning=int(d['gcm_binning']),
            phase_binning=int(d['phase_binning']),
            use_molecular_signatures=bool(d['use_molecular_signatures']),
            use_continuum_stellar=bool(d.get('use_continuum_stellar',cls._defaults['use_stellar_continuum'])),
            nmax=int(d['nmax']),
            lmax=int(d['lmax']),
            continuum=list(d['continuum']),
        )




import numpy as np
from astropy import units as u, constants as c
from VSPEC import config


class ForwardSpectra:
    """
    Spectra from a forward model.

    Parameters
    ----------
    func : callable
        The function that performs the evaluation of the spectra.

    """

    def __init__(self, func):
        """
        pass a function that does the real work
        """
        self.evaluate = func

    @classmethod
    def blackbody(cls):
        """
        Create a ForwardSpectra object for blackbody spectra.

        Returns
        -------
        ForwardSpectra
            A ForwardSpectra object for blackbody spectra.

        """

        def func(wl: u.Quantity, teff: u.Quantity):
            """
            Evaluate the blackbody spectra at the given wavelength and temperature.

            Parameters
            ----------
            wl : jax.numpy.ndarray
                The array of wavelength values in micrometers.
            teff : float
                The temperature in Kelvin.

            Returns
            -------
            jax.numpy.ndarray
                The flux values of the blackbody spectra at the given wavelength and temperature.

            """
            if not wl.unit.physical_type == u.um.physical_type:
                raise u.UnitTypeError('wl is wrong physical type')
            if not teff.unit.physical_type == u.K.physical_type:
                raise u.UnitTypeError('teff is wrong physical type')
            A = 2 * c.h * c.c**2/wl**5
            B = np.exp(((c.h*c.c)/(wl*c.k_B*teff)
                        ).to(u.dimensionless_unscaled)) - 1
            return (A/B).to(config.flux_unit)
        return cls(func)



"""VSPEC main module

This module performs all of VSPEC's interaction with the user.
It contains the `ObservationModel` class, which has methods that
perform all of the model aggregation from the rest of the package
and PSG.
"""

from pathlib import Path
import warnings
import logging
from functools import partial
from typing import Dict, List, Tuple

import numpy as np
from astropy import units as u
from astropy.table import QTable
from tqdm.auto import tqdm

from GridPolator import GridSpectra
from GridPolator.binning import get_wavelengths

import pypsg
from pypsg import docker as psg_docker

import vspec_vsm as vsm

# from VSPEC import variable_star_model as vsm
from VSPEC.config import PSG_CFG_MAX_LINES, N_ZFILL
from VSPEC import config
from VSPEC.geometry import SystemGeometry
from VSPEC.helpers import isclose, arrange_teff
from VSPEC.helpers import check_and_build_dir, get_filename
from VSPEC.helpers import get_planet_indicies
from VSPEC.psg_api import get_reflected
from VSPEC.psg_api import change_psg_parameters
from VSPEC.params.read import InternalParameters
from VSPEC.spectra import ForwardSpectra


class ObservationModel:
    """
    Main class that stores the information of this simulation.

    Parameters
    ----------
    params : VSPEC.params.InternalParameters
        The global parameters describing the VSPEC simulation.

    Examples
    --------
    >>> model = ObservationModel.from_yaml('config.yaml')
    >>> model.build_planet()
    Build Planet: 100%|██████████| 26/26 [00:02<00:00,  8.68it/s]
    >>> model.build_spectra()
    Build Spectra: 100%|██████████| 26/26 [00:15<00:00,  1.66it/s]

    >>> model = ObservationModel(params=my_params)
    >>> model.build_planet()
    Build Planet: 100%|██████████| 18/18 [00:07<00:00,  2.62it/s]
    >>> model.build_spectra()
    Build Spectra: 100%|██████████| 18/18 [02:35<00:00,  8.65s/it]

    """

    def __init__(
        self,
        params: InternalParameters
    ):
        if isinstance(params, (Path, str)):
            msg = 'Please use the `from_yaml` classmethod'
            raise TypeError(msg)
        self.params = params
        self.verbose = params.header.verbose
        self._build_directories()
        self.rng = np.random.default_rng(self.params.header.seed)
        # Load later when needed.
        self.spec: GridSpectra | ForwardSpectra = None
        self.star: vsm.Star = None
        self.bb = ForwardSpectra.blackbody()
        self.logger = logging.Logger('VSPEC')
        self.logger.setLevel(logging.DEBUG)
        fh = logging.FileHandler(self.directories['parent'] / 'psg.log',mode='w')
        fh.setLevel(logging.DEBUG)
        self.logger.addHandler(fh)

    @classmethod
    def from_yaml(cls, config_path: Path):
        """
        Initialize a VSPEC run from a YAML file.

        Parameters
        ----------
        config_path : pathlib.Path
            The path to the YAML file.
        """
        params = InternalParameters.from_yaml(config_path)
        return cls(params)

    class __flags__:
        """
        Flags attribute

        Attributes
        ----------
        psg_needs_set : bool
            For security reasons, PSG config files can contain a
            maximum of ~2000 lines. After that number, the server
            will no longer update after our API call. In this case,
            we issue the `set` command. Unfortuantely, we must re-
            upload the GCM.
        """
        psg_needs_set = True
    _directories = {
        'parent': '',
        'all_model': 'AllModelSpectraValues',
        'psg_combined': 'PSGCombinedSpectra',
        'psg_thermal': 'PSGThermalSpectra',
        'psg_noise': 'PSGNoise',
        'psg_layers': 'PSGLayers',
        'psg_configs': 'PSGConfig'
    }

    @property
    def directories(self) -> dict:
        """
        The directory structure for the VSPEC run.

        Returns
        -------
        dict
            Keys represent the identifiers of directories, and the values are
            `pathlib.Path` objects.

        """
        parent_dir = self.params.header.data_path
        dir_dict = {key: parent_dir/value for key,
                    value in self._directories.items()}
        return dir_dict

    def _wrap_iterator(self, iterator, **kwargs):
        """
        Wrapper for iterators so that `tqdm` can be used
        only if `self.verbose` > 0

        Parameters
        ----------
        iterator : iterable
            Iterator to be passed to `tqdm`
        **kwargs : dict
            The keywords to pass to `tqdm`

        Returns
        -------
        iterable
            The iterator wrapped appropriately.
        """
        if self.verbose > 0:
            return tqdm(iterator, **kwargs)
        else:
            return iterator

    def _build_directories(self):
        """
        Build the file system for this run.
        """
        for _, path in self.directories.items():
            check_and_build_dir(path)

    def _load_spectra(self):
        """
        Load a GridSpec instance.

        Returns
        -------
        VSPEC.spectra.GridSpec
            The spectal grid object to draw stellar spectra from.
        """
        if self.params.star.spectral_grid == 'default':
            teffs = arrange_teff(
                self.params.header.teff_min,
                self.params.header.teff_max
            )
            spec = GridSpectra.from_vspec(
                w1=self.params.inst.bandpass.wl_blue,
                w2=self.params.inst.bandpass.wl_red,
                resolving_power=self.params.inst.bandpass.resolving_power,
                teffs=teffs
            )
            return spec
        elif self.params.star.spectral_grid == 'bb':
            return ForwardSpectra.blackbody()

    @property
    def _wl(self):
        """
        The wavelength axis of the observation.

        Returns
        -------
        wl : astropy.units.Quantity
            The wavelength axis.
        """
        return get_wavelengths(
            resolving_power=self.params.inst.bandpass.resolving_power,
            lam1=self.params.inst.bandpass.wl_blue.to_value(config.wl_unit),
            lam2=self.params.inst.bandpass.wl_red.to_value(config.wl_unit)
        )[:-1]*config.wl_unit

    def _get_model_spectrum(self, teff: u.Quantity) -> u.Quantity:
        """
        Get the interpolated spectrum given an effective temperature.

        Parameters
        ----------
        teff : astropy.units.Quantity
            The effective temperature

        Returns
        -------
        astropy.units.Quantity
            The flux of the spectrum.

        Notes
        -----
        This function applies the solid angle correction.
        """
        if self.spec is None:
            self.spec = self._load_spectra()

        match self.params.star.spectral_grid:
            case 'default':
                teffs = np.atleast_1d(teff.to_value(config.teff_unit))
                return self.spec.evaluate(
                    params=(teffs,),
                    wl=np.array(self._wl.to_value(config.wl_unit))
                )[0, :] * config.flux_unit * self.params.flux_correction
            case 'bb':
                self.spec: ForwardSpectra
                return self.spec.evaluate(self._wl, teff) * self.params.flux_correction
        raise ValueError(
            f'Spectral grid {self.params.star.spectral_grid} not recognized')

    def get_observation_parameters(self) -> SystemGeometry:
        """

        Get an object to store and compute the geometric observational
        parameters for this simulation.


        Returns
        -------
        VSPEC.geometry.SystemGeometry
            An bbject storing the geometric observation parameters
            of this simulation.
        """
        return SystemGeometry(inclination=self.params.system.inclination,
                              init_stellar_lon=0*u.deg,
                              init_planet_phase=self.params.planet.init_phase,
                              stellar_period=self.params.star.period,
                              orbital_period=self.params.planet.orbit_period,
                              semimajor_axis=self.params.planet.semimajor_axis,
                              planetary_rot_period=self.params.planet.rotation_period,
                              planetary_init_substellar_lon=self.params.planet.init_substellar_lon,
                              stellar_offset_amp=self.params.star.misalignment,
                              stellar_offset_phase=self.params.star.misalignment_dir,
                              eccentricity=self.params.planet.eccentricity,
                              phase_of_periasteron=self.params.system.phase_of_periasteron,
                              system_distance=self.params.system.distance,
                              obliquity=self.params.planet.obliquity,
                              obliquity_direction=self.params.planet.obliquity_direction)

    def _get_observation_plan(self, observation_parameters: SystemGeometry, planet=False) -> QTable:
        """
        Compute the locations and geometries of each object in this simulation.

        Parameters
        ----------
        observation_parameters : VSPEC.geometry.SystemGeometry
            An object containting the system geometry.
        planet : bool
            If true, use the planet phase binning parameter to
            compute the number of steps.

        Returns
        -------
        QTable
            A table containing the observation plan.
        """
        int_time = self.params.obs.integration_time
        obs_time = self.params.obs.observation_time

        if not planet:
            n_obs = int(
                round((obs_time / int_time).to_value(u.dimensionless_unscaled)))
            start_times = np.arange(0, n_obs) * int_time
        else:
            n_obs = int(round(
                (obs_time / int_time/self.params.psg.phase_binning).to_value(u.dimensionless_unscaled)))
            start_times = np.arange(0, n_obs+1) * \
                int_time * self.params.psg.phase_binning

        return observation_parameters.get_observation_plan(start_times)

    def _check_psg(self):
        """
        Check that PSG is configured correctly.

        Raises
        ------
        RuntimeError
            If a local PSG container is specified but the port is not in use
            (i.e. PSG is not running).

        Warns
        -----
        RuntimeWarning
            If calling the online PSG API, but no API key is specified.
        """
        if pypsg.settings.get_setting('url') == pypsg.settings.PSG_URL:
            if pypsg.settings.get_setting('api_key') is None:
                warnings.warn('PSG is being called without an API key. ' +
                              'After 100 API calls in a 24hr period you will need to get a key. ' +
                              'We suggest installing PSG locally using docker. (see https://psg.gsfc.nasa.gov/help.php#handbook)',
                              RuntimeWarning)
        else:
            if not psg_docker.is_psg_installed():
                msg = 'PSG is not installed. '
                msg += 'We suggest installing PSG locally using docker. (see https://psg.gsfc.nasa.gov/help.php#handbook) '
                msg += 'otherwise set the `pypsg` url setting to the online PSG url:\n'
                msg += f'pypsg.settings.save_settings(url=\'{pypsg.settings.PSG_URL}\')'
                raise RuntimeError(msg)
            elif not psg_docker.is_psg_running():
                raise RuntimeError('PSG is not running.\n' +
                                   'Type `docker start psg` in the command line ' +
                                   'or run `pypsg.docker.start_psg()`')

    def _upload_gcm(self, obstime: u.Quantity = 0*u.s, update=False):
        """
        Upload GCM file to PSG

        Parameters
        ----------
        is_ncdf : bool
            Whether or not to use a netCDF file for the GCM
        obstime : astropy.units.Quantity, default=0*u.s
            The time since the start of the observation.
        update : bool
            Whether to use the `'upd'` keyword rather than `'set'`
        """
        if not self.params.gcm.is_staic:
            kwargs = {'obs_time': obstime}
        else:
            kwargs = {}
        cfg = self.params.gcm.to_pycfg(**kwargs)
        caller = pypsg.APICall(
            cfg=cfg,
            output_type='upd' if update else 'set',
            app='globes',
            logger=self.logger
        )
        _ = caller()
        if not update:
            self.__flags__.psg_needs_set = False

    def _set_static_config(self):
        """
        Upload the non-changing parts of the PSG config.
        """
        cfg = self.params.to_pyconfig()
        caller = pypsg.APICall(
            cfg=cfg,
            output_type='upd',
            app='globes',
            logger=self.logger
        )
        _ = caller()

    def _psg_all_call(
        self,
        phase: u.Quantity,
        orbit_radius_coeff: float,
        sub_stellar_lon: u.Quantity,
        sub_stellar_lat: u.Quantity,
        pl_sub_obs_lon: u.Quantity,
        pl_sub_obs_lat: u.Quantity,
        include_star: bool,
        path_dict: dict,
        i: int
    ):
        """
        Update the PSG config with time-dependent values.

        Parameters
        ----------
        phase : astropy.units.Quantity
            The phase of the planet
        orbit_radius_coeff : float
            The ratio between the current planet-host difference
            and the semimajor axis
        sub_stellar_lon : astropy.units.Quantity
            The sub-stellar longitude of the planet.
        sub_stellar_lat : astropy.units.Quantity
            The sub-stellar latitude of the planet.
        pl_sub_obs_lon : astropy.units.Quantity
            The sub-observer longitude of the planet.
        pl_sub_obs_lat : astropy.units.Quantity
            The sub-observer latitude of the planet.
        include_star : bool
            Whether to include the star in the simulation.
        path_dict : dict
            A dictionary that determines where each downloaded file
            gets written to.
        i : int
            The index of the current observation
        """
        cfg = change_psg_parameters(
            params=self.params,
            phase=phase,
            orbit_radius_coeff=orbit_radius_coeff,
            sub_stellar_lon=sub_stellar_lon,
            sub_stellar_lat=sub_stellar_lat,
            pl_sub_obs_lon=pl_sub_obs_lon,
            pl_sub_obs_lat=pl_sub_obs_lat,
            include_star=include_star
        )
        caller = pypsg.APICall(
            cfg=cfg,
            output_type='all',
            app='globes',
            logger=self.logger
        )
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            response = caller()
            for _w in w:
                if _w.category is pypsg.cfg.ConfigTooLongWarning:
                    self.__flags__.psg_needs_set = True
        rad = response.rad
        noi = response.noi
        lyr = response.lyr
        cfg = response.cfg
        
        if rad is None:
            raise RuntimeError('PSG returned no .rad file')
        else:
            key = 'rad'
            filename = get_filename(i, N_ZFILL, 'fits')
            rad.write(
                path_dict[key]/filename,
                format='fits',
                overwrite=True
            )
            if include_star and 'Stellar' not in rad.colnames:
                raise ValueError('No Column named "Stellar" in .rad file')

        if noi is None or 'noi' not in path_dict:
            pass
        else:
            key = 'noi'
            filename = get_filename(i, N_ZFILL, 'fits')
            noi.write(
                path_dict[key]/filename,
                format='fits',
                overwrite=True
            )

        if cfg is None:
            raise RuntimeError('PSG returned no .cfg file')
        elif 'cfg' not in path_dict:
            pass
        else:
            key = 'cfg'
            filename = get_filename(i, N_ZFILL, 'cfg')
            cfg.to_file(path_dict[key]/filename)

            # TODO: Check config against expected

        if lyr is None or 'lyr' not in path_dict:
            pass
        else:
            key = 'lyr'
            filename = get_filename(i, N_ZFILL, 'fits')
            lyr.to_fits(path_dict[key]/filename)

    def _check_config(self, cfg_from_psg: pypsg.PyConfig):
        """
        Validate the config file recieved from PSG to ensure that
        the parameters sent are the parameters used in the simulation.

        Parameters
        ----------
        cfg_from_psg : pypsg.PyConfig
            The config file recieved from PSG.

        Raises
        ------
        RuntimeError
            If the config recieved does not match the config sent.
        """
        n_lines = len(cfg_from_psg.content.split('\n'))
        if n_lines > PSG_CFG_MAX_LINES:
            self.__flags__.psg_needs_set = True

        expected_cfg = self.params.to_pyconfig()

        assert expected_cfg.target == cfg_from_psg.target, 'PSG target does not match'
        assert expected_cfg.geometry == cfg_from_psg.geometry, 'PSG geometry does not match'
        assert expected_cfg.atmosphere == cfg_from_psg.atmosphere, 'PSG atmosphere does not match'
        assert expected_cfg.surface == cfg_from_psg.surface, 'PSG surface does not match'
        assert expected_cfg.generator == cfg_from_psg.generator, 'PSG generator does not match'
        assert expected_cfg.telescope == cfg_from_psg.telescope, 'PSG telescope does not match'
        assert expected_cfg.noise == cfg_from_psg.noise, 'PSG noise does not match'
        # do not check the GCM


    def build_planet(self):
        """
        Use the PSG GlobES API to construct a planetary phase curve.
        Follow steps in original PlanetBuilder.py file
        
        Examples
        --------
        >>> model = ObservationModel.from_yaml(CFG_PATH)
        >>> model.build_planet()
        Build Planet: 100%|██████████| 36/36 [00:07<00:00,  4.60it/s]

        """
        # check that psg is running
        self._check_psg()
        # for not using globes, append all configurations instead of rewritting

        ####################################
        # Initial upload of GCM

        self._upload_gcm(
            obstime=0*u.s,
            update=False
        )
        ####################################
        # Set observation parameters that do not change
        self._set_static_config()

        ####################################
        # Calculate observation parameters
        observation_parameters = self.get_observation_parameters()
        obs_plan = self._get_observation_plan(
            observation_parameters, planet=True)

        obs_info_filename = Path(
            self.directories['parent']) / 'observation_info.fits'
        obs_plan.write(obs_info_filename, overwrite=True)

        if self.verbose > 0:
            print(
                f'Starting at phase {self.params.planet.init_phase}, observe for {self.params.obs.observation_time} in {self.params.planet_total_images} steps')
            print('Phases = ' +
                  str(np.round(np.asarray((obs_plan['phase']/u.deg).to(u.Unit(''))), 2)) + ' deg')
        ####################################
        # iterate through phases
        for i in self._wrap_iterator(range(self.params.planet_total_images+1), desc='Build Planet', total=self.params.planet_total_images):
            phase = obs_plan['phase'][i]
            sub_stellar_lon = obs_plan['sub_stellar_lon'][i]
            sub_stellar_lat = obs_plan['sub_stellar_lat'][i]
            pl_sub_obs_lon = obs_plan['planet_sub_obs_lon'][i]
            pl_sub_obs_lat = obs_plan['planet_sub_obs_lat'][i]
            orbit_radius_coeff = obs_plan['orbit_radius'][i]
            obs_time = obs_plan['time'][i] - obs_plan['time'][0]

            if (not self.params.gcm.is_staic) or self.__flags__.psg_needs_set:
                # enter if we need a reset or if there is time dependence
                upload = partial(self._upload_gcm, obstime=obs_time)
                if self.__flags__.psg_needs_set:  # do a reset if needed
                    upload(update=False)
                    self._set_static_config()
                    self.__flags__.psg_needs_set = False
                else:  # update if it's just time dependence.
                    upload(update=True)

            # Write updates to the config to change the phase value and ensure the star is of type 'StarType'
            update_config = partial(
                self._psg_all_call,
                phase=phase,
                orbit_radius_coeff=orbit_radius_coeff,
                sub_stellar_lon=sub_stellar_lon,
                sub_stellar_lat=sub_stellar_lat,
                pl_sub_obs_lon=pl_sub_obs_lon,
                pl_sub_obs_lat=pl_sub_obs_lat
            )
            
            path_dict = {
                'rad': Path(self.directories['psg_combined']),
                'noi': Path(self.directories['psg_noise']),
                'cfg': Path(self.directories['psg_configs']),
            }
            update_config(
                include_star=True,
                path_dict=path_dict,
                i=i
            )
            # write updates to config file to remove star flux
            
            path_dict = {
                'rad': Path(self.directories['psg_thermal']),
                'lyr': Path(self.directories['psg_layers'])
            }
            update_config(
                include_star=False,
                path_dict=path_dict,
                i=i
            )

    def _build_star(self):
        """
        Build a variable star model based on user-specified parameters.
        """
        self.star = self.params.star.to_star(
            rng=self.rng,
            seed=self.params.header.seed
        )

    def _warm_up_star(self, spot_warmup_time: u.Quantity = 0*u.day, facula_warmup_time: u.Quantity = 0*u.day):
        """
        "Warm up" the star. Generate spots, faculae, and/or flares for the star.
        The goal is to approach growth-decay equillibrium, something that is hard to
        do with a purely "hot star" method (like 
        `VPSEC.variable_star_model.Star.generate_mature_spots`).

        Parameters
        ----------
        spot_warm_up_time : astropy.units.Quantity [time], default=0*u.day
            The time to run to approach spot equillibrium.
        facula_warmup_time : astropy.units.Quantity [time], default=0*u.hr
            The time to run to approach faculae equillibrium.
        """
        if self.params.star.spots.initial_coverage > 0.0:
            self.star.generate_mature_spots(
                self.params.star.spots.initial_coverage)
            print(f'Generated {len(self.star.spots.spots)} mature spots')
        spot_warm_up_step = 1*u.day
        facula_warm_up_step = 1*u.hr
        N_steps_spot = int(
            round((spot_warmup_time/spot_warm_up_step).to(u.Unit('')).value))
        N_steps_facula = int(
            round((facula_warmup_time/facula_warm_up_step).to(u.Unit('')).value))
        if N_steps_spot > 0:
            for _ in self._wrap_iterator(range(N_steps_spot), desc='Spot Warmup', total=N_steps_spot):
                self.star.birth_spots(spot_warm_up_step)
                self.star.age(spot_warm_up_step)
        if N_steps_facula > 0:
            for _ in self._wrap_iterator(range(N_steps_facula), desc='Facula Warmup', total=N_steps_facula):
                self.star.birth_faculae(facula_warm_up_step)
                self.star.age(facula_warm_up_step)

        self.star.get_flares_over_observation(
            self.params.obs.observation_time)

    def _calculate_composite_stellar_spectrum(
        self,
        sub_obs_coords,
        tstart,
        tfinish,
        granulation_fraction: float = 0.0,
        orbit_radius: u.Quantity = 1*u.AU,
        planet_radius: u.Quantity = 1*u.R_earth,
        phase: u.Quantity = 90*u.deg,
        inclination: u.Quantity = 0*u.deg,
        transit_depth: np.ndarray | float = 0
    ):
        """
        Compute the stellar spectrum given an integration window and the
        side of the star facing the observer.

        Parameters
        ----------
        sub_obs_coords : dict
            A dictionary containing stellar sub-observer coordinates.
        tstart : astropy.units.Quantity [time]
            The starting time of the observation.
        tfinish : astropy.units.Quantity [time]
            The ending time of the observation.
        granulation_fraction : float
            The fraction of the quiet photosphere that has a lower Teff due to granulation

        Returns
        -------
        base_wave : astropy.units.Quantity [wavelength]
            The wavelength coordinates of the stellar spectrum.
        base_flux : astropy.units.Quantity [flambda]
            The composite stellar flux

        Raises
        ------
        ValueError
            If wavenelength coordinates do not match.
        """
        total, covered, pl_frac = self.star.calc_coverage(
            sub_obs_coords,
            granulation_fraction=granulation_fraction,
            orbit_radius=orbit_radius,
            planet_radius=planet_radius,
            phase=phase,
            inclination=inclination
        )
        visible_flares = self.star.get_flare_int_over_timeperiod(
            tstart, tfinish, sub_obs_coords)
        base_flux = self._get_model_spectrum(self.params.star.teff)
        base_flux = base_flux * 0
        # add up star flux before considering transit
        for teff, coverage in total.items():
            if coverage > 0:
                flux = self._get_model_spectrum(teff)
                if not flux.shape == base_flux.shape:
                    raise ValueError('All arrays must have same shape.')
                base_flux = base_flux + flux * coverage
        # get flux of transited region
        transit_flux = base_flux*0
        for teff, coverage in covered.items():
            if coverage > 0:
                flux = self._get_model_spectrum(teff)
                if not flux.shape == base_flux.shape:
                    raise ValueError('All arrays must have same shape.')
                transit_flux = transit_flux + flux * coverage
        # scale according to effective radius
        transit_flux = transit_flux * transit_depth
        base_flux = base_flux - transit_flux
        # add in flares
        for flare in visible_flares:
            teff = flare['Teff']
            timearea = flare['timearea']
            eff_area = (timearea/(tfinish-tstart)).to(u.km**2)
            correction = (eff_area/self.params.system.distance **
                          2).to_value(u.dimensionless_unscaled)
            flux = self.bb.evaluate(self._wl, teff) * correction
            base_flux = base_flux + flux

        return base_flux.to(config.flux_unit), pl_frac

    def _calculate_reflected_spectra(
        self,
        N1: int,
        N2: int,
        N1_frac: float,
        sub_planet_flux: u.Quantity,
        pl_frac: float
    ) -> u.Quantity:
        """
        Calculate the reflected spectrum based on PSG output and
        our own stellar model. We scale the reflected spectra from PSG
        to our model.

        Parameters
        ----------
        N1 : int
            The planet index immediately before the current epoch.
        N2 : int
            The planet index immediately after the current epoch.
        N1_frac : float
            The fraction of the `N1` epoch to use in interpolation.
        sub_planet_flux : astropy.units.Quantity
            Stellar flux to scale to.
        pl_frac : float
            The fraction of the planet that is visible (in case of
            eclipse)

        Returns
        -------
        reflected_flux : astropy.units.Quantity
            Reflected flux.

        Raises
        ------
        ValueError
            If the PSG flux unit code is not recognized.
        ValueError
            If the wavelength coordinates from the loaded spectra do not match.
        """
        psg_combined_path1 = Path(
            self.directories['psg_combined']) / get_filename(N1, N_ZFILL, 'fits')
        psg_thermal_path1 = Path(
            self.directories['psg_thermal']) / get_filename(N1, N_ZFILL, 'fits')
        psg_combined_path2 = Path(
            self.directories['psg_combined']) / get_filename(N2, N_ZFILL, 'fits')
        psg_thermal_path2 = Path(
            self.directories['psg_thermal']) / get_filename(N2, N_ZFILL, 'fits')

        reflected = []

        for psg_combined_path, psg_thermal_path in zip([psg_combined_path1, psg_combined_path2],
                                                       [psg_thermal_path1, psg_thermal_path2]):
            combined = pypsg.PyRad.read(psg_combined_path, format='fits')
            thermal = pypsg.PyRad.read(psg_thermal_path, format='fits')

            planet_reflection_only = get_reflected(
                combined,
                thermal,
                self.params.planet.name
            )
            planet_reflection_fraction: np.ndarray = (
                planet_reflection_only / combined['Stellar']).to_value(u.dimensionless_unscaled)

            planet_reflection_adj = sub_planet_flux * planet_reflection_fraction
            reflected.append(planet_reflection_adj)

        return reflected[0] * N1_frac + reflected[1] * (1-N1_frac)*pl_frac

    def _get_transit(
        self,
        N1: int,
        N2: int,
        N1_frac: float,
        phase: u.Quantity,
        orbit_radius: u.Quantity
    ) -> Tuple[u.Quantity, np.ndarray]:
        """
        Get the transit spectra calculated by PSG

        Parameters
        ----------
        N1 : int
            The planet index immediately before the current epoch.
        N2 : int
            The planet index immediately after the current epoch.
        N1_frac : float
            The fraction of the `N1` epoch to use in interpolation.
        pl_frac : float
            The fraction of the planet that is visible (not eclipsed)

        Returns
        -------
        wavelength : astropy.units.Quantity
            The wavelength of the thermal emission.
        flux : np.ndarray
            The dimensionless fraction of light blocked by the planet at each wavelength.

        Raises
        ------
        ValueError
            If the PSG flux unit code is not recognized.
        ValueError
            If the wavelength coordinates from the loaded spectra do not match.
        """
        psg_cmb_path1 = Path(
            self.directories['psg_combined']) / get_filename(N1, N_ZFILL, 'fits')
        psg_cmb_path2 = Path(
            self.directories['psg_combined']) / get_filename(N2, N_ZFILL, 'fits')

        wavelength = []
        transit: List[np.ndarray] = []

        for psg_cmb_path in [psg_cmb_path1, psg_cmb_path2]:
            cmb_rad: pypsg.PyRad = pypsg.PyRad.read(
                psg_cmb_path, format='fits')

            wavelength.append(cmb_rad.wl)
            if 'Transit' in cmb_rad.colnames:
                tran = cmb_rad['Transit']
                star = cmb_rad['Stellar']
                blocked_frac = -1 * \
                    (tran/star).to_value(u.dimensionless_unscaled)
                transit.append(blocked_frac)
            else:
                star = cmb_rad['Stellar']
                blocked_frac = 0*(star/star).to_value(u.dimensionless_unscaled)
                transit.append(blocked_frac)

        if not np.all(isclose(wavelength[0], wavelength[1], 1e-3*u.um)):
            raise ValueError('The wavelength coordinates must be equivalent.')

        frac_absorbed: np.ndarray = transit[0]*N1_frac + transit[1]*(1-N1_frac)

        depth_bare_rock: float = (self.params.planet.radius /
                                  self.params.star.radius).to_value(u.dimensionless_unscaled)**2
        return wavelength[0], frac_absorbed/depth_bare_rock

        # pl_frac_visible = self.star.get_pl_frac(
        #     angle_past_midtransit=phase+180*u.deg,
        #     orbit_radius=orbit_radius,
        #     planet_radius=self.params.planet.radius,
        #     inclination=self.params.system.inclination
        # )
        # pl_frac_covering = 1-pl_frac_visible * (self.params.planet.radius/self.params.star.radius).to_value(u.dimensionless_unscaled)**2
        # if pl_frac_covering == 0:
        #     normalized_frac_absorbed = pl_frac_covering*0
        # else:
        #     normalized_frac_absorbed = frac_absorbed/pl_frac_covering/depth_bare_rock
        # return wavelength[0], normalized_frac_absorbed

    def _calculate_noise(self, N1: int, N2: int, N1_frac: float, time_scale_factor: float, cmb_flux):
        """
        Calculate the noise in our model based on the noise output from PSG.

        Parameters
        ----------
        N1 : int
            The planet index immediately before the current epoch.
        N2 : int
            The planet index immediately after the current epoch.
        N1_frac : float
            The fraction of the `N1` epoch to use in interpolation.
        time_scale_factor : float
            A scaling factor to apply to the noise at the end of the calculation.
            This is 1 if the planet and star sampling has the same cadence. Otherwise,
            it is usually `sqrt(self.planet_phase_binning)`.
        cmb_flux : astropy.units.Quantity [flambda]
            The flux of the combined spectrum.

        Returns
        -------
        noise : astropy.units.Quantity [flambda]
            The noise in our model.

        Raises
        ------
        ValueError
            If the PSG flux unit code is not recognized.
        ValueError
            If the wavelength coordinates from the loaded spectra do not match.
        """
        psg_combined_path1 = Path(
            self.directories['psg_combined']) / get_filename(N1, N_ZFILL, 'fits')
        psg_noise_path1 = Path(
            self.directories['psg_noise']) / get_filename(N1, N_ZFILL, 'fits')
        psg_combined_path2 = Path(
            self.directories['psg_combined']) / get_filename(N2, N_ZFILL, 'fits')
        psg_noise_path2 = Path(
            self.directories['psg_noise']) / get_filename(N2, N_ZFILL, 'fits')

        psg_noise_source = []
        psg_source = []

        for psg_combined_path, psg_noise_path in zip(
            [psg_combined_path1, psg_combined_path2],
            [psg_noise_path1, psg_noise_path2]
        ):
            combined: pypsg.PyRad = pypsg.PyRad.read(
                psg_combined_path, format='fits')
            noise: pypsg.PyRad = pypsg.PyRad.read(
                psg_noise_path, format='fits')
            psg_noise_source.append(noise['Source'])
            psg_source.append(combined['Total'])

        psg_noise_source = psg_noise_source[0] * \
            N1_frac + psg_noise_source[1] * (1-N1_frac)
        psg_source = psg_source[0]*N1_frac + psg_source[1] * (1-N1_frac)

        model_noise = psg_noise_source * np.sqrt(cmb_flux/psg_source)
        noise_sq = (model_noise**2
                    + (noise['Detector'])**2
                    + (noise['Telescope'])**2
                    + (noise['Background'])**2)
        return np.sqrt(noise_sq) * time_scale_factor

    def _get_thermal_spectrum(self, N1: int, N2: int, N1_frac: float, pl_frac: float):
        """
        Get the thermal emission spectra calculated by PSG

        Parameters
        ----------
        N1 : int
            The planet index immediately before the current epoch.
        N2 : int
            The planet index immediately after the current epoch.
        N1_frac : float
            The fraction of the `N1` epoch to use in interpolation.
        pl_frac : float
            The fraction of the planet that is visible (not eclipsed)

        Returns
        -------
        wavelength : astropy.units.Quantity [wavelength]
            The wavelength of the thermal emission.
        flux : astropy.units.Quantity [flambda]
            The flux of the thermal emission.

        Raises
        ------
        ValueError
            If the PSG flux unit code is not recognized.
        ValueError
            If the wavelength coordinates from the loaded spectra do not match.
        """
        psg_thermal_path1 = Path(
            self.directories['psg_thermal']) / get_filename(N1, N_ZFILL, 'fits')
        psg_thermal_path2 = Path(
            self.directories['psg_thermal']) / get_filename(N2, N_ZFILL, 'fits')

        wavelength = []
        thermal = []

        for psg_thermal_path in [psg_thermal_path1, psg_thermal_path2]:
            thermal_rad: pypsg.PyRad = pypsg.PyRad.read(
                psg_thermal_path, format='fits')

            wavelength.append(thermal_rad.wl)
            try:
                planet_name = self.params.planet.name.replace(' ', '-')
                thermal.append(thermal_rad[planet_name])
            except KeyError:
                try:
                    thermal.append(thermal_rad['Thermal'])
                except KeyError as err2:  # There is something wrong with the planet name. See issue #25
                    msg = f'Neither "{planet_name}" nor "Thermal" were found in {thermal_rad.colnames}. '
                    new_issue_url = 'https://github.com/VSPEC-collab/VSPEC/issues/new'
                    msg += f'If this is a valid planet name, please report this bug at {new_issue_url}'
                    raise RuntimeError(msg) from err2

        if not np.all(isclose(wavelength[0], wavelength[1], 1e-3*u.um)):
            raise ValueError('The wavelength coordinates must be equivalent.')

        return thermal[0]*N1_frac + thermal[1]*(1-N1_frac)*pl_frac

    def _get_layer_data(self, N1: int, N2: int, N1_frac: float) -> pypsg.PyLyr:
        """
        Interpolate between two PSG .lyr files.

        Parameters
        ----------
        N1 : int
            The planet index immediately before the current epoch.
        N2 : int
            The planet index immediately after the current epoch.
        N1_frac : float
            The fraction of the `N1` epoch to use in interpolation.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the interpolated layer data.

        Raises
        ------
        ValueError
            If the layer file columns of layer numbers do not match.
        """
        psg_layers_path1 = Path(
            self.directories['psg_layers']) / get_filename(N1, N_ZFILL, 'fits')
        psg_layers_path2 = Path(
            self.directories['psg_layers']) / get_filename(N2, N_ZFILL, 'fits')
        layers1 = pypsg.PyLyr.from_fits(psg_layers_path1)
        layers2 = pypsg.PyLyr.from_fits(psg_layers_path2)

        if not np.all(layers1.prof.colnames == layers2.prof.colnames):
            raise ValueError(
                'Layer Profiles table files must have matching columns')
        if not np.all(layers1.cg.colnames == layers2.cg.colnames):
            raise ValueError(
                'Layer CG table files must have matching columns')

        prof_names = layers1.prof.colnames
        cg_names = layers1.cg.colnames

        prof_data = {}
        cg_data = {}
        for name in prof_names:
            prof_data[name] = layers1.prof[name] * \
                N1_frac + layers2.prof[name] * (1-N1_frac)
        for name in cg_names:
            cg_data[name] = layers1.cg[name] * \
                N1_frac + layers2.cg[name] * (1-N1_frac)

        return pypsg.PyLyr(
            prof=QTable(data=prof_data, meta=layers1.prof.meta),
            cg=QTable(data=cg_data, meta=layers1.cg.meta)
        )

    def build_spectra(self):
        """
        Integrate our stellar model with PSG to produce a variable
        host + planet simulation.
        Follow the original Build_Spectra.py file to construct phase curve outputs.
        
        Examples
        --------
        >>> model = ObservationModel.from_yaml(CFG_PATH) # build_planet() has been run previously
        >>> model.build_spectra()
        Build Spectra: 100%|██████████| 36/36 [00:23<00:00,  1.53it/s]

        """
        if self.star is None:  # user can define a custom star before calling this function, e.g. for a specific spot pattern
            self._build_star()
            self._warm_up_star(spot_warmup_time=self.params.star.spots.burn_in,
                              facula_warmup_time=self.params.star.faculae.burn_in)
        observation_parameters = self.get_observation_parameters()
        observation_info = self._get_observation_plan(
            observation_parameters, planet=False)
        # write observation info to file
        obs_info_filename = Path(
            self.directories['all_model']) / 'observation_info.fits'
        observation_info.write(obs_info_filename, overwrite=True)

        planet_observation_info = self._get_observation_plan(
            observation_parameters, planet=True)
        planet_times = planet_observation_info['time']

        time_step = self.params.obs.integration_time
        planet_time_step = self.params.obs.observation_time / \
            (self.params.planet_total_images+1)
        granulation_fractions = self.star.get_granulation_coverage(
            observation_info['time'])

        for index in self._wrap_iterator(range(self.params.obs.total_images), desc='Build Spectra', total=self.params.obs.total_images, position=0, leave=True):
            tindex = observation_info['time'][index]
            tstart = tindex - observation_info['time'][0]
            tfinish = tstart + time_step
            planet_phase = observation_info['phase'][index]
            sub_obs_lon = observation_info['sub_obs_lon'][index]
            sub_obs_lat = observation_info['sub_obs_lat'][index]
            orbital_radius = observation_info['orbit_radius'][index] * \
                self.params.planet.semimajor_axis
            granulation_fraction = granulation_fractions[index]
            N1, N2 = get_planet_indicies(planet_times, tindex)
            N1_frac = (planet_times[N2] - tindex)/planet_time_step
            N1_frac = N1_frac.to_value(u.dimensionless_unscaled)

            sub_planet_lon = observation_info['sub_planet_lon'][index]
            sub_planet_lat = observation_info['sub_planet_lat'][index]

            _, transit_depth = self._get_transit(
                N1, N2, N1_frac, planet_phase, orbital_radius)

            comp_flux, pl_frac = self._calculate_composite_stellar_spectrum(
                {'lat': sub_obs_lat, 'lon': sub_obs_lon}, tstart, tfinish,
                granulation_fraction=granulation_fraction,
                orbit_radius=orbital_radius,
                planet_radius=self.params.planet.radius,
                phase=planet_phase,
                inclination=self.params.system.inclination,
                transit_depth=transit_depth
            )
            true_star, _ = self._calculate_composite_stellar_spectrum(
                {'lat': sub_obs_lat, 'lon': sub_obs_lon}, tstart, tfinish,
                granulation_fraction=granulation_fraction,
                orbit_radius=orbital_radius,
                planet_radius=self.params.planet.radius,
                phase=planet_phase,
                inclination=self.params.system.inclination,
                transit_depth=0.
            )
            to_planet_flux, _ = self._calculate_composite_stellar_spectrum(
                {'lat': sub_planet_lat, 'lon': sub_planet_lon}, tstart, tfinish,
                granulation_fraction=granulation_fraction
            )

            reflection_flux_adj = self._calculate_reflected_spectra(
                N1, N2, N1_frac, to_planet_flux, pl_frac)

            thermal_spectrum = self._get_thermal_spectrum(
                N1, N2, N1_frac, pl_frac)

            combined_flux = comp_flux + reflection_flux_adj + thermal_spectrum

            noise_flux_adj = self._calculate_noise(N1, N2, N1_frac,
                                                  np.sqrt(
                                                      (planet_time_step/time_step).to_value(u.dimensionless_unscaled)),
                                                  combined_flux)
            wl = self._wl
            data: Dict[str, u.Quantity] = {}
            data['wavelength'] = wl
            data['star'] = true_star
            data['star_towards_planet'] = to_planet_flux
            data['reflected'] = reflection_flux_adj
            data['planet_thermal'] = thermal_spectrum
            data['total'] = combined_flux
            data['noise'] = noise_flux_adj

            df = QTable(data)

            outfile = self.directories['all_model'] / \
                get_filename(index, N_ZFILL, 'fits')
            df.write(outfile, overwrite=True)

            # layers
            if self.params.psg.use_molecular_signatures:
                layerdat = self._get_layer_data(N1, N2, N1_frac)
                outfile = self.directories['all_model'] / \
                    f'layer{str(index).zfill(N_ZFILL)}.fits'
                layerdat.to_fits(outfile)

            self.star.birth_spots(time_step)
            self.star.birth_faculae(time_step)
            self.star.age(time_step)



"""
VSPEC: Variable Star Phase Curve

VSPEC is designed as an observation simulation suite for rocky
exoplanets orbiting variable stars. Built for the Mid-IR Exoplanet
CLimate Explorer (MIRECLE) mission concept, this package is flexible
enough to simulate observations over a wide range of conditions.

VSPEC utilizes the Planetary Spectrum Generator (PSG, psg.gsfc.nasa.gov)
to perform radiative transfer calculations
"""


__author__ = 'Cameron Kelahan, Ted Johnson'
__version__ = '0.6.0'

from .main import ObservationModel
from .analysis import PhaseAnalyzer
from . import params



"""
Stellar Parameters
"""
from typing import Union, Tuple
from astropy import units as u
import numpy as np
import yaml

from vspec_vsm import (
    SpotGenerator,
    FaculaGenerator,
    FlareGenerator,
    Granulation,
    Star,
    SpotCollection,
    FaculaCollection,
    CoordinateGrid
)
from vspec_vsm import config as vsm_config


from VSPEC.config import MSH, PRESET_PATH
from VSPEC.params.base import BaseParameters


class LimbDarkeningParameters(BaseParameters):
    """
    Limb Darkening Parameters for the Quadratic
    Limb Darkening Law

    Parameters
    ----------
    u1 : float
        The limb darkening parameter u1
    u2 : float
        The limb darkening parameter u2

    Attributes
    ----------
    u1 : float
        The limb darkening parameter u1
    u2 : float
        The limb darkening parameter u2

    Notes
    -----
    Quadratic Law (Kopal, 1950, cited in :cite:t:`2022AJ....163..228P`)

    .. math::

        \\frac{I(\\mu)}{I(1)} = 1 - u_1 (1-\\mu) - u_2 (1-\\mu)^2

    Examples
    --------
    >>> params_dict = {'preset': 'solar'}
    >>> params = LimbDarkeningParameters.from_dict(params_dict)

    In the example above, the 'solar' preset configuration is used to create an instance
    of LimbDarkeningParameters.

    >>> params_dict = {'u1': 0.3, 'u2': 0.1}
    >>> params = LimbDarkeningParameters.from_dict(params_dict)

    In the example above, custom values for 'u1' and 'u2' are provided to create an instance
    of LimbDarkeningParameters.

    """

    def __init__(self, u1: float, u2: float):
        self.u1 = u1
        self.u2 = u2

    @classmethod
    def _from_dict(cls, d):
        return cls(
            float(d['u1']),
            float(d['u2'])
        )

    @classmethod
    def solar(cls):
        """
        From :cite:t:`2013A&A...552A..16C`
        S4 bandpass
        """
        return cls(0.0473, 0.0841)

    @classmethod
    def proxima(cls):
        """
        From :cite:t:`2012A&A...546A..14C`
        S4 bandpass
        """
        return cls(0.0551, 0.1075)

    @classmethod
    def trappist(cls):
        """
        From :cite:t:`2012A&A...546A..14C`
        S4 bandpass
        """
        return cls(0.0778, 0.1619)

    @classmethod
    def lambertian(cls):
        """
        No limb darkening.
        """
        return cls(0., 0.)


class SpotParameters(BaseParameters):
    """
    Parameters controling variability from star spots.

    Parameters
    ----------
    distribution : str
        The distribution function to be used for the spot positions.
        Either ``iso`` for an isotropic distribution or ``solar`` for two bands at :math:`\\pm 15^\\circ` latitude.
    initial_coverage : float
        The spot coverage created initially by generating spots at random stages of life.
    area_mean : astropy.units.Quantity
        The mean area of a spot on the star's surface is MSH.
    area_logsigma : float
        The standard deviation of the spot areas. This is a lognormal
        distribution, so the units of this value are dex.
    teff_umbra : astropy.units.Quantity
        The effective temperature of the spot umbrae.
    teff_penumbra : astropy.units.Quantity
        The effective temperature of the spot penumbrae.
    equillibrium_coverage : float
        The fractional coverage of the star's surface by spots. This is the value
        at growth-decay equillibrium.
    burn_in : astropy.units.Quantity
        The duration of the burn-in period, during which the spot coverage approaches
        equillibrium.
    growth_rate : astropy.units.Quantity
        The rate at which new spots grow.
    decay_rate : astropy.units.Quantity
        The rate at which existing spots decay.
    initial_area : astropy.units.Quantity
        The initial area of newly created spots.

    Attributes
    ----------
    distribution : str
        The distribution function to be used for the spot positions. 'iso' or 'solar'.
    initial_coverage : float
        The coverage for a 'hot start'.
    area_logmean : float
        The mean area of a spot on the star's surface is MSH.
    area_logsigma : float
        The standard deviation of the spot areas. This is a lognormal
        distribution, so the units of this value are dex.
    teff_umbra : astropy.units.Quantity
        The effective temperature of the spot umbrae.
    teff_penumbra : astropy.units.Quantity
        The effective temperature of the spot penumbrae.
    equillibrium_coverage : float
        The fractional coverage of the star's surface by spots. This is the value
        at growth-decay equillibrium, and different from the 'hot start' value given
        by `initial_coverage`.
    burn_in : astropy.units.Quantity
        The duration of the burn-in period, during which the spot coverage approaches
        equillibrium.
    growth_rate : astropy.units.Quantity
        The rate at which new spots grow.
    decay_rate : astropy.units.Quantity
        The rate at which existing spots decay.
    initial_area : astropy.units.Quantity
        The initial area of newly created spots.

    """
    _PRESET_PATH = PRESET_PATH / 'spots.yaml'
    """
    The path to the preset file.
    """

    def __init__(
        self,
        distribution: str,
        initial_coverage: float,
        area_mean: u.Quantity,
        area_logsigma: float,
        teff_umbra: u.Quantity,
        teff_penumbra: u.Quantity,
        equillibrium_coverage: float,
        burn_in: u.Quantity,
        growth_rate: u.Quantity,
        decay_rate: u.Quantity,
        initial_area: u.Quantity
    ):
        self.distribution = distribution
        self.initial_coverage = initial_coverage
        self.area_mean = area_mean
        self.area_logsigma = area_logsigma
        self.teff_umbra = teff_umbra
        self.teff_penumbra = teff_penumbra
        self.equillibrium_coverage = equillibrium_coverage
        self.burn_in = burn_in
        self.growth_rate = growth_rate
        self.decay_rate = decay_rate
        self.initial_area = initial_area
        self._validate()

    def _validate(self):
        """
        Validate class instance
        """
        if self.distribution not in ['solar', 'iso']:
            raise ValueError('`distribution` must either be `solar` or `iso`')
        if self.initial_coverage > 1 or self.initial_coverage < 0:
            raise ValueError('`initial_coverage` must be between 0 and 1.')
        if self.equillibrium_coverage > 1 or self.equillibrium_coverage < 0:
            raise ValueError(
                '`equillibrium_coverage` must be between 0 and 1.')

    @classmethod
    def _from_dict(cls, d):
        u.add_enabled_units([MSH])
        return cls(
            distribution=str(d['distribution']),
            initial_coverage=float(d['initial_coverage']),
            area_mean=u.Quantity(d['area_mean']),
            area_logsigma=float(d['area_logsigma']),
            teff_umbra=u.Quantity(d['teff_umbra']),
            teff_penumbra=u.Quantity(d['teff_penumbra']),
            equillibrium_coverage=float(d['equillibrium_coverage']),
            burn_in=u.Quantity(d['burn_in']),
            growth_rate=u.Quantity(d['growth_rate']),
            decay_rate=u.Quantity(d['decay_rate']),
            initial_area=u.Quantity(d['initial_area'])
        )

    @classmethod
    def from_preset(cls, name):
        """
        Load a ``SpotParameters`` instance from a preset file.

        Parameters
        ----------
        name : str
            The name of the preset to load.

        Returns
        -------
        SpotParameters
            The class instance loaded from a preset.
        """
        with open(cls._PRESET_PATH, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data[name])

    @classmethod
    def none(cls):
        """
        No spots
        """
        return cls.from_preset('none')

    @classmethod
    def mdwarf(cls):
        """
        Static Spots
        """
        return cls.from_preset('mdwarf')

    @classmethod
    def solar(cls):
        """
        Solar-style spots
        """
        return cls.from_preset('solar')

    def to_generator(
        self,
        grid_params: Union[int, Tuple[int, int]] = (
            vsm_config.NLAT, vsm_config.NLON),
        gridmaker: CoordinateGrid = None,
        rng: np.random.Generator = np.random.default_rng()
    ) -> SpotGenerator:
        """
        Create a `vspec_vsm.SpotGenerator` instance from the class instance.

        Parameters
        ----------
        grid_params : Union[int, Tuple[int, int]], optional
            If tuple, the number of grid points in the latitude and longitude directions.
            If int, the number of total grid points for a sprial grid.
            Defaults to (vsm_config.NLAT, vsm_config.NLON).
        gridmaker : CoordinateGrid, optional
            An instance of `vspec_vsm.CoordinateGrid` to use for the grid.
            Defaults to None.
        rng : np.random.Generator, optional
            The random number generator to use. Defaults to np.random.default_rng().

        Returns
        -------
        vspec_vsm.SpotGenerator
            The `vspec_vsm.SpotGenerator` instance.
        """
        return SpotGenerator(
            dist_area_mean=self.area_mean,
            dist_area_logsigma=self.area_logsigma,
            umbra_teff=self.teff_umbra,
            penumbra_teff=self.teff_penumbra,
            growth_rate=self.growth_rate,
            decay_rate=self.decay_rate,
            init_area=vsm_config.starspot_initial_area,
            distribution=self.distribution,
            coverage=self.equillibrium_coverage,
            grid_params=grid_params,
            gridmaker=gridmaker,
            rng=rng
        )


class FaculaParameters(BaseParameters):
    """
    Facula Parameters

    Parameters
    ----------
    distribution : str
        The distribution used to generate the faculae on the star.
        Currently only 'iso' is supported.
    equillibrium_coverage : float
        The fraction of the star's surface covered by the faculae at growth-decay equilibrium.
    burn_in : astropy.units.Quantity
        The duration of the burn-in period, during which the facula coverage approaches
        equillibrium.
    mean_radius : astropy.units.Quantity
        The mean radius of the faculae.
    logsigma_radius : float
        The standard deviation of :math:`\\log{r_{0}}`.
    depth : astropy.units.Quantity
        The depth of the facula depression.
    mean_timescale : astropy.units.Quantity
        The mean faculae lifetime.
    logsigma_timescale : float
        The standard deviation of :math:`\\log{\\tau}`
    floor_teff_slope : astropy.units.Quantity
        The slope of the radius-Teff relationship
    floor_teff_min_rad : astropy.units.Quantity
        The minimum radius at which the floor is visible. Otherwise the facula
        is a bright point -- even near disk center.
    floor_teff_base_dteff : astropy.units.Quantity
        The Teff of the floor at the minimum radius.
    wall_teff_slope : astropy.units.Quantity
        The slope of the radius-Teff relationship
    wall_teff_intercept : astropy.units.Quantity
        The Teff of the wall when :math:`R = 0`.

    Attributes
    ----------
    distribution : str
        The distribution used to generate the faculae on the star.
    equillibrium_coverage : float
        The fraction of the star's surface covered by the faculae at growth-decay equilibrium.
    burn_in : astropy.units.Quantity
        The duration of the burn-in period, during which the facula coverage approaches
        equillibrium.
    mean_radius : astropy.units.Quantity
        The mean radius of the faculae.
    logsigma_radius : float
        The standard deviation of :math:`\\log{r_{0}}`.
    depth : astropy.units.Quantity
        The depth of the facula depression.
    mean_timescale : astropy.units.Quantity
        The mean faculae lifetime.
    logsigma_timescale : float
        The standard deviation of :math:`\\log{\\tau}`
    floor_teff_slope : astropy.units.Quantity
        The slope of the radius-Teff relationship
    floor_teff_min_rad : astropy.units.Quantity
        The minimum radius at which the floor is visible. Otherwise the facula
        is a bright point -- even near disk center.
    floor_teff_base_dteff : astropy.units.Quantity
        The Teff of the floor at the minimum radius.
    wall_teff_slope : astropy.units.Quantity
        The slope of the radius-Teff relationship
    wall_teff_intercept : astropy.units.Quantity
        The Teff of the wall when :math:`R = 0`.
    """
    _PRESET_PATH = PRESET_PATH / 'faculae.yaml'

    def __init__(
        self,
        distribution: str,
        equillibrium_coverage: float,
        burn_in: u.Quantity,
        mean_radius: u.Quantity,
        logsigma_radius: float,
        depth: u.Quantity,
        mean_timescale: u.Quantity,
        logsigma_timescale: float,
        floor_teff_slope: u.Quantity,
        floor_teff_min_rad: u.Quantity,
        floor_teff_base_dteff: u.Quantity,
        wall_teff_slope: u.Quantity,
        wall_teff_intercept: u.Quantity,
    ):
        self.distribution = distribution
        self.equillibrium_coverage = equillibrium_coverage
        self.burn_in = burn_in
        self.mean_radius = mean_radius
        self.logsigma_radius = logsigma_radius
        self.depth = depth
        self.mean_timescale = mean_timescale
        self.logsigma_timescale = logsigma_timescale
        self.floor_teff_slope = floor_teff_slope
        self.floor_teff_min_rad = floor_teff_min_rad
        self.floor_teff_base_dteff = floor_teff_base_dteff
        self.wall_teff_slope = wall_teff_slope
        self.wall_teff_intercept = wall_teff_intercept
        self._validate()

    def _validate(self):
        """
        Validate class instance
        """
        if self.distribution not in ['iso']:
            raise ValueError('`distribution` must be `iso`')
        if self.equillibrium_coverage > 1 or self.equillibrium_coverage < 0:
            raise ValueError(
                '`equillibrium_coverage` must be between 0 and 1.')

    @classmethod
    def _from_dict(cls, d):
        return cls(
            distribution=str(d['distribution']),
            equillibrium_coverage=float(d['equillibrium_coverage']),
            burn_in=u.Quantity(d['burn_in']),
            mean_radius=u.Quantity(d['mean_radius']),
            logsigma_radius=float(d['logsigma_radius']),
            mean_timescale=u.Quantity(d['mean_timescale']),
            logsigma_timescale=float(d['logsigma_timescale']),
            depth=u.Quantity(d['depth']),
            floor_teff_slope=u.Quantity(d['floor_teff_slope']),
            floor_teff_min_rad=u.Quantity(d['floor_teff_min_rad']),
            floor_teff_base_dteff=u.Quantity(d['floor_teff_base_dteff']),
            wall_teff_slope=u.Quantity(d['wall_teff_slope']),
            wall_teff_intercept=u.Quantity(d['wall_teff_intercept'])
        )

    @classmethod
    def from_preset(cls, name):
        """
        Load a ``FaculaParameters`` instance from a preset file.

        Parameters
        ----------
        name : str
            The name of the preset to load.

        Returns
        -------
        FaculaParameters
            The class instance loaded from a preset.
        """
        return super().from_preset(name)

    @classmethod
    def none(cls):
        """
        Load a parameter set without faculae.
        """
        return cls.from_preset('none')

    @classmethod
    def std(cls):
        """
        Load a parameter preset with simple standard faculae
        for testing.
        """
        return cls.from_preset('std')

    def to_generator(
        self,
        grid_params: Union[int, Tuple[int, int]] = (
            vsm_config.NLAT, vsm_config.NLON),
        gridmaker: CoordinateGrid = None,
        rng: np.random.Generator = np.random.default_rng()
    ) -> FaculaGenerator:
        """
        Construct a `vspec_vsm.FaculaGenerator` instance from the class.

        Parameters
        ----------
        grid_params : Union[int,Tuple[int, int]]
            If tuple, the number of grid points in the latitude and longitude
            directions. If int, the number of total grid points for a sprial
            grid. Defaults to (vsm_config.NLAT, vsm_config.NLON).
        gridmaker : CoordinateGrid
            An instance of `vspec_vsm.CoordinateGrid` to use for the grid.
            Defaults to None.
        rng : np.random.Generator
            Random number generator. Defaults to np.random.default_rng().

        Returns
        -------
        vspec_vsm.FaculaGenerator
            The `vspec_vsm.FaculaGenerator` instance.
        """
        return FaculaGenerator(
            dist_r_peak=self.mean_radius,
            dist_r_logsigma=self.logsigma_radius,
            depth=self.depth,
            dist_life_peak=self.mean_timescale,
            dist_life_logsigma=self.logsigma_timescale,
            floor_teff_slope=self.floor_teff_slope,
            floor_teff_min_rad=self.floor_teff_min_rad,
            floor_teff_base_dteff=self.floor_teff_base_dteff,
            wall_teff_slope=self.wall_teff_slope,
            wall_teff_intercept=self.wall_teff_intercept,
            coverage=self.equillibrium_coverage,
            dist=self.distribution,
            grid_params=grid_params,
            gridmaker=gridmaker,
            rng=rng
        )


class FlareParameters(BaseParameters):
    """
    Class to store stellar flare parameters

    Parameters
    ----------
    dist_teff_mean : astropy.units.Quantity
        The mean of the temperature distribution.
    dist_teff_sigma : astropy.units.Quantity
        The standard deviation of the temperature distribution.
    dist_fwhm_mean : astropy.units.Quantity
        The mean of the FWHM distribution.
    dist_fwhm_logsigma : float
        The standard deviation of the logorithm of the FWHM distribution in dex.
    alpha : float
        The slope of the log frequency - log energy relationship.
    beta : float
        The y-intercept of the log frequency - log energy relationship.
    min_energy : astropy.units.Quantity
        The minimum energy to consider. Set to ``np.inf*u.erg`` to disable flares.
    cluster_size : int
        The typical size of flare clusters.

    Attributes
    ----------
    dist_teff_mean : astropy.units.Quantity
        The mean of the temperature distribution.
    dist_teff_sigma : astropy.units.Quantity
        The standard deviation of the temperature distribution.
    dist_fwhm_mean : astropy.units.Quantity
        The mean of the FWHM distribution.
    dist_fwhm_logsigma : float
        The standard deviation of the logorithm of the FWHM distribution in dex.
    alpha : float
        The slope of the log frequency - log energy relationship.
    beta : float
        The y-intercept of the log frequency - log energy relationship.
    min_energy : astropy.units.Quantity
        The minimum energy to consider. Set to ``np.inf*u.erg`` to disable flares.
    cluster_size : int
        The typical size of flare clusters.
    """
    _PRESET_PATH = PRESET_PATH / 'flares.yaml'

    def __init__(
        self,
        dist_teff_mean: u.Quantity,
        dist_teff_sigma: u.Quantity,
        dist_fwhm_mean: u.Quantity,
        dist_fwhm_logsigma: float,
        alpha: float,
        beta: float,
        min_energy: u.Quantity,
        cluster_size: int
    ):
        self.dist_teff_mean = dist_teff_mean
        self.dist_teff_sigma = dist_teff_sigma
        self.dist_fwhm_mean = dist_fwhm_mean
        self.dist_fwhm_logsigma = dist_fwhm_logsigma
        self.alpha = alpha
        self.beta = beta
        self.min_energy = min_energy
        self.cluster_size = cluster_size

    @classmethod
    def from_preset(cls, name):
        """
        Load a ``FlareParameters`` configuration from a preset.

        Parameters
        ----------
        name : str
            The name of the preset to load.

        Returns
        -------
        FlareParameters
            The class instance loaded from a preset.
        """
        return super().from_preset(name)

    @classmethod
    def _from_dict(cls, d):
        return cls(
            dist_teff_mean=u.Quantity(d['dist_teff_mean']),
            dist_teff_sigma=u.Quantity(d['dist_teff_sigma']),
            dist_fwhm_mean=u.Quantity(d['dist_fwhm_mean']),
            dist_fwhm_logsigma=float(d['dist_fwhm_logsigma']),
            alpha=float(d['alpha']),
            beta=float(d['beta']),
            min_energy=u.Quantity(d['min_energy']),
            cluster_size=int(d['cluster_size'])
        )

    @classmethod
    def none(cls):
        """
        A configuration with no flares.
        """
        return cls.from_preset('none')

    @classmethod
    def std(cls):
        """
        A standard flare configuration for testing.
        """
        return cls.from_preset('std')

    def to_generator(
        self,
        rng: np.random.Generator = np.random.default_rng()
    ) -> FlareGenerator:
        """
        Create a `vspec_vsm.FlareGenerator` instance from the class instance.

        Parameters
        ----------
        rng : np.random.Generator, optional
            The random number generator to use. Defaults to np.random.default_rng().

        Returns
        -------
        vspec_vsm.FlareGenerator
            The `vspec_vsm.FlareGenerator` instance.
        """
        return FlareGenerator(
            dist_teff_mean=self.dist_teff_mean,
            dist_teff_sigma=self.dist_teff_sigma,
            dist_fwhm_mean=self.dist_fwhm_mean,
            dist_fwhm_logsigma=self.dist_fwhm_logsigma,
            alpha=self.alpha,
            beta=self.beta,
            min_energy=self.min_energy,
            cluster_size=self.cluster_size,
            rng=rng
        )


class GranulationParameters(BaseParameters):
    """
    Granulation Parameters

    Parameters
    ----------
    mean : float
        The mean coverage of low-teff granulation.
    amp : float
        The amplitude of granulation oscillations.
    period : astropy.units.Quantity
        The period of granulation oscillations.
    dteff : astropy.units.Quantity
        The difference between the quiet photosphere and the low-teff granulation region.

    Attributes
    ----------
    mean : float
        The mean coverage of low-teff granulation.
    amp : float
        The amplitude of granulation oscillations.
    period : astropy.units.Quantity
        The period of granulation oscillations.
    dteff : astropy.units.Quantity
        The difference between the quiet photosphere and the low-teff granulation region.
    """

    def __init__(
        self,
        mean: float,
        amp: float,
        period: u.Quantity,
        dteff: u.Quantity
    ):
        self.mean = mean
        self.amp = amp
        self.period = period
        self.dteff = dteff
        self.validate()

    def validate(self):
        if self.mean > 1 or self.mean < 0:
            raise ValueError('`mean` must be between 0 and 1.')
        if self.amp > 1 or self.amp < 0:
            raise ValueError('`amp` must be between 0 and 1.')

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            mean=float(d['mean']),
            amp=float(d['amp']),
            period=u.Quantity(d['period']),
            dteff=u.Quantity(d['dteff'])
        )
    
    @classmethod
    def from_preset(cls, name):
        match name:
            case 'none':
                return cls.none()
            case 'std':
                return cls.std()
            case _:
                return super().from_preset(name)

    @classmethod
    def std(cls):
        return cls(
            0.2, 0.01, 5*u.day, 200*u.K
        )

    @classmethod
    def none(cls):
        return cls(
            0.0, 0.00, 5*u.day, 200*u.K
        )

    def to_generator(
        self,
        seed: int = 0
    ) -> Granulation:
        """
        Create a `vspec_vsm.Granulation` instance from the class instance.

        Parameters
        ----------
        seed : int, optional
            The seed for the random number generator. Defaults to 0.

        Returns
        -------
        vspec_vsm.Granulation
            The `vspec_vsm.Granulation` instance.
        """
        return Granulation(
            mean_coverage=self.mean,
            amplitude=self.amp,
            period=self.period,
            dteff=self.dteff,
            seed=seed
        )


class StarParameters(BaseParameters):
    """
    Parameters describing the ``VSPEC`` stellar model.

    Parameters
    ----------
    psg_star_template : str
        The template used for the stellar model.
    teff : astropy.units.Quantity
        The effective temperature of the star.
    mass : astropy.units.Quantity
        The mass of the star.
    radius : astropy.units.Quantity
        The radius of the star.
    period : astropy.units.Quantity
        The rotational period of the star.
    misalignment : astropy.units.Quantity
        The misalignment between the stellar rotation axis and the orbital axis.
    misalignment_dir : astropy.units.Quantity
        The direction of stellar rotation axis misalignment.
    ld : LimbDarkeningParameters
        The limb darkening parameters of the star.
    spots : SpotParameters
        The parameters of the spots on the star.
    faculae : FaculaParameters
        The parameters of the faculae on the star.
    flares : FlareParameters
        The parameters of the flares on the star.
    granulation : GranulationParameters
        The parameters of the granulation on the star.
    grid_params : tuple or int
        The parameters for the grid. If tuple, we use a rectangular
        grid (Nlat, Nlon). If int, we use a spiral grid with
        `grid_params` points evenly spaced.
    spectral_grid : str
        The grid of stellar spectra to use. Currently only
        ``default`` and ``bb`` are supported. Use ``default``
        in general, but if you want efficiency and only care about
        the absolute spectrum of the planet, ``bb`` is recommended
        as it can cut the runtime in half.

    Attributes
    ----------
    psg_star_template : str
        The template used for the stellar model.
    teff : astropy.units.Quantity
        The effective temperature of the star.
    mass : astropy.units.Quantity
        The mass of the star.
    radius : astropy.units.Quantity
        The radius of the star.
    period : astropy.units.Quantity
        The rotational period of the star.
    misalignment : astropy.units.Quantity
        The misalignment between the stellar rotation axis and the orbital axis.
    misalignment_dir : astropy.units.Quantity
        The direction of stellar rotation axis misalignment.
    ld : LimbDarkeningParameters
        The limb darkening parameters of the star.
    spots : SpotParameters
        The parameters of the spots on the star.
    faculae : FaculaParameters
        The parameters of the faculae on the star.
    flares : FlareParameters
        The parameters of the flares on the star.
    granulation : GranulationParameters
        The parameters of the granulation on the star.
    grid_params : tuple or int
        The parameters for the grid.
    spectral_grid : str
        The grid of stellar spectra to use.
    """

    def __init__(
        self,
        psg_star_template: str,
        teff: u.Quantity,
        mass: u.Quantity,
        radius: u.Quantity,
        period: u.Quantity,
        misalignment: u.Quantity,
        misalignment_dir: u.Quantity,
        ld: LimbDarkeningParameters,
        spots: SpotParameters,
        faculae: FaculaParameters,
        flares: FlareParameters,
        granulation: GranulationParameters,
        grid_params: tuple,
        spectral_grid: str
    ):
        self.psg_star_template = psg_star_template
        self.teff = teff
        self.mass = mass
        self.radius = radius
        self.period = period
        self.misalignment = misalignment
        self.misalignment_dir = misalignment_dir
        self.ld = ld
        self.spots = spots
        self.faculae = faculae
        self.flares = flares
        self.granulation = granulation
        self.grid_params = grid_params
        self.spectral_grid = spectral_grid

    @classmethod
    def from_dict(cls, d: dict):
        """
        Construct a ``StarParameters`` object from a dictionary.

        Parameters
        ----------
        d : dict
            The dictionary representing the ``StarParameters`` object.

        Notes
        -----
        Available presets include ``static_proxima``, ``spotted_proxima``, ``flaring_proxima``, and ``proxima``.
        """
        return super().from_dict(d)

    @classmethod
    def _from_dict(cls, d: dict):
        try:
            grid_params = tuple(d['grid_params'])
        except TypeError:
            grid_params = int(d['grid_params'])
        return cls(
            psg_star_template=str(d['psg_star_template']),
            teff=u.Quantity(d['teff']),
            mass=u.Quantity(d['mass']),
            radius=u.Quantity(d['radius']),
            period=u.Quantity(d['period']),
            misalignment=u.Quantity(d['misalignment']),
            misalignment_dir=u.Quantity(d['misalignment_dir']),
            ld=LimbDarkeningParameters.from_dict(d['ld']),
            spots=SpotParameters.from_dict(d['spots']),
            faculae=FaculaParameters.from_dict(d['faculae']),
            flares=FlareParameters.from_dict(d['flares']),
            granulation=GranulationParameters.from_dict(d['granulation']),
            grid_params=grid_params,
            spectral_grid=str(d['spectral_grid'])
        )

    @classmethod
    def from_preset(cls, name):
        match name:
            case 'static_proxima':
                return cls.static_proxima()
            case 'spotted_proxima':
                return cls.spotted_proxima()
            case 'flaring_proxima':
                return cls.flaring_proxima()
            case 'proxima':
                return cls.proxima()
            case _:
                raise ValueError(f'Unknown preset: {name}')    
    @classmethod
    def static_proxima(cls):
        """
        A Proxima Centauri-like star that has no variability.
        """
        return cls(
            psg_star_template='M',
            teff=3300*u.K,
            mass=0.12*u.M_sun,
            radius=0.154*u.R_sun,
            period=40*u.day,
            misalignment=0*u.deg,
            misalignment_dir=0*u.deg,
            ld=LimbDarkeningParameters.proxima(),
            spots=SpotParameters.none(),
            faculae=FaculaParameters.none(),
            flares=FlareParameters.none(),
            granulation=GranulationParameters.none(),
            grid_params=(500, 1000),
            spectral_grid='default'
        )

    @classmethod
    def spotted_proxima(cls):
        """
        A Proxima Centauri-like star that has spots.
        """
        return cls(
            psg_star_template='M',
            teff=3300*u.K,
            mass=0.12*u.M_sun,
            radius=0.154*u.R_sun,
            period=40*u.day,
            misalignment=0*u.deg,
            misalignment_dir=0*u.deg,
            ld=LimbDarkeningParameters.proxima(),
            spots=SpotParameters.mdwarf(),
            faculae=FaculaParameters.none(),
            flares=FlareParameters.none(),
            granulation=GranulationParameters.none(),
            grid_params=(500, 1000),
            spectral_grid='default'
        )

    @classmethod
    def flaring_proxima(cls):
        """
        A Proxima Centauri-like star that flares.
        """
        return cls(
            psg_star_template='M',
            teff=3300*u.K,
            mass=0.12*u.M_sun,
            radius=0.154*u.R_sun,
            period=40*u.day,
            misalignment=0*u.deg,
            misalignment_dir=0*u.deg,
            ld=LimbDarkeningParameters.proxima(),
            spots=SpotParameters.none(),
            faculae=FaculaParameters.none(),
            flares=FlareParameters.std(),
            granulation=GranulationParameters.none(),
            grid_params=(500, 1000),
            spectral_grid='default'
        )

    @classmethod
    def proxima(cls):
        """
        A Proxima Centauri-like star that has spots and flares.
        """
        return cls(
            psg_star_template='M',
            teff=3300*u.K,
            mass=0.12*u.M_sun,
            radius=0.154*u.R_sun,
            period=40*u.day,
            misalignment=0*u.deg,
            misalignment_dir=0*u.deg,
            ld=LimbDarkeningParameters.proxima(),
            spots=SpotParameters.mdwarf(),
            faculae=FaculaParameters.none(),
            flares=FlareParameters.std(),
            granulation=GranulationParameters.std(),
            grid_params=(500, 1000),
            spectral_grid='default'
        )

    def to_star(
        self,
        rng: np.random.Generator = np.random.default_rng(),
        seed: int = 0
    ) -> Star:
        """
        Create a `vspec_vsm.Star` instance from the class instance.

        Parameters
        ----------
        rng : np.random.Generator, optional
            The random number generator to use. Defaults to np.random.default_rng().
        seed : int, optional
            The seed for the random number generator. Defaults to 0.
        """
        return Star(
            radius=self.radius,
            period=self.period,
            teff=self.teff,
            spots=SpotCollection(grid_params=self.grid_params),
            faculae=FaculaCollection(grid_params=self.grid_params),
            grid_params=self.grid_params,
            flare_generator=self.flares.to_generator(rng=rng),
            spot_generator=self.spots.to_generator(
                grid_params=self.grid_params,
                gridmaker=None,
                rng=rng
            ),
            fac_generator=self.faculae.to_generator(
                grid_params=self.grid_params,
                gridmaker=None,
                rng=rng
            ),
            granulation=self.granulation.to_generator(seed=seed),
            u1=self.ld.u1,
            u2=self.ld.u2,
            rng=rng
        )



"""
Physics of heat transport


Based on :cite:t:`2011ApJ...726...82C`

"""

from typing import Tuple
from astropy import units as u, constants as c
import numpy as np
from scipy.integrate import solve_ivp, solve_bvp
import warnings
from copy import deepcopy
from scipy.interpolate import interp1d

from pypsg.globes import PyGCM
from pypsg.globes import structure


def get_flux(
    teff_star: u.Quantity,
    r_star: u.Quantity,
    r_orbit: u.Quantity
) -> u.Quantity:
    """
    Get the flux of a star at a certain orbital radius

    Parameters
    ----------
    teff_star : astropy.units.Quantity
        The effective temperature of the star.
    r_star : astropy.units.Quantity
        The radius of the star.
    r_orbit : astropy.units.Quantity
        The orbital radius.

    Returns
    -------
    flux : astropy.units.Quantity
        The flux of the star at the planet.

    """
    # pylint: disable-next:no-member
    flux = c.sigma_sb * teff_star**4 * (r_star/r_orbit)**2
    return flux.to(u.Unit('W m-2'))


def get_psi(lon: u.Quantity):
    """
    Translate between longitude and Psi.

    Since the math we are doing for the energy balance requires
    longituide on [-180,180] deg, we cast everything to this range.

    Parameters
    ----------
    lon : astropy.units.Quantity
        The longituide to cast.

    Returns
    -------
    np.ndarray
        The longitude coordinate in radians cast to [-pi,pi]
    """
    unit = u.deg
    alon = np.atleast_1d(lon.to_value(unit))
    return (np.where(alon <= 180, alon, alon-360)*unit).to_value(u.rad)


def pcos(x: u.Quantity):
    """
    'Positive cosine' -- :math:`\\max{\\cos{x},0}`

    Parameters
    ----------
    x : astropy.units.Quantity
        The angle to take the cosine of.

    Returns
    -------
    np.ndarray
        The maximum of 0 and cosine `x`.
    """
    cos = np.cos(x)
    if isinstance(cos, u.Quantity):
        cos = cos.to_value(u.dimensionless_unscaled)
    return np.where(cos > 0, cos, 0)


def colat(lat: u.Quantity):
    """
    Get the colatitude.

    Parameters
    ----------
    lat : astropy.units.Quantity
        The latitude.

    Returns
    -------
    colat : astropy.units.Quantity
        The colatitude.
    """
    return 90*u.deg - lat


def get_t0(
    star_teff: u.Quantity,
    albedo: float,
    r_star: u.Quantity,
    r_orbit: u.Quantity
):
    """
    Get the fiducial temperature for a planet's atmosphere, as
    defined by :cite:t:`2011ApJ...726...82C`, equation 4.

    Parameters
    ----------
    teff_star : astropy.units.Quantity
        The effective temperature of the star.
    albedo : float
        The Bond albedo of the planet.
    r_star : astropy.units.Quantity
        The radius of the star.
    r_orbit : astropy.units.Quantity
        The orbital radius.

    Returns
    -------
    t0: astropy.units.Quantity
        The fiducial temperature of the planet.
    """
    return (star_teff * (1-albedo)**0.25 * np.sqrt(r_star/r_orbit)).to(u.K)


class EquatorSolverWarning(RuntimeWarning):
    """
    Warning to tell the user to use a different solver.
    """


def get_equator_curve(epsilon: float, n_points: int, mode: str = 'ivp_reflect'):
    """
    Get the temperature along the equator given thermal
    inertia `epsilon`. This is computed by integrating
    equation 10 of :cite:t:`2011ApJ...726...82C`.

    Parameters
    ----------
    epsilon : float
        The thermal inertial of the planet.
    n_points : int
        The numer of longitude points to return in the final array.
    mode : str
        The method to use to find the solution. This can be one of
        'ivp_reflect', 'bvp', 'ivp_iterate', 'analytic'.

    Returns
    -------
    lons : np.ndarray
        The longitude points in radians, starting at -pi
    tsurf : np.ndarray
        The unitless ratio between the surface temperature and the fiducial temperature.

    Warns
    -----
    RuntimeWarning
        If the specified `mode` is not valid with the specified `epsilon`.

    Notes
    -----
    Of the four solving methods, each has a region in which it is valid. Some are optimized for
    `epsilon` ~ 1, some for small `epsilon`, and some for when `epsilon` is very large.

    `ivp_reflect`: Integrate from pi to -pi, to find an initial condition to integrate
    from -pi to pi. For small `epsilon`, this method is very robust. Valid for `epsilon` < 1

    `bvp`: Solve the boundary condition problem for T(-pi) = T(pi). Valid for `epsilon` > 1

    `ivp_interate`: Integrate the ODE until the boundaries are within 1%. Valid for `epsilon` < 0.5

    `analytic`: Use an analytic approximation. Very fast, valid for `epsilon` > 10

    """
    def get_lons(n):
        return np.linspace(-np.pi, np.pi, n)

    def func(phi, temperature):
        with np.errstate(over='ignore'):
            return (pcos(phi) - temperature**4)/epsilon

    def minus_func(phi, temperature):
        with np.errstate(over='ignore'):
            return -(pcos(phi) - temperature**4)/epsilon
    if mode == 'ivp_reflect':
        if epsilon > 1:
            msg = 'Using method `ivp_reflect` with `epsilon` > 1. Energy balance may be invalid.'
            warnings.warn(msg, EquatorSolverWarning)
        lons = get_lons(n_points)
        y0 = np.atleast_1d(1)
        result = solve_ivp(minus_func, (np.pi, -np.pi), y0=y0)
        y_final = np.atleast_1d(np.mean(result.y[:, -1]))

        result = solve_ivp(func, (-np.pi, np.pi), y0=y_final, t_eval=lons)
        if not result.success:
            raise ValueError(result.message)
        return result.t, result.y.T
    elif mode == 'bvp':
        if epsilon < 1:
            msg = 'Using method `bvp` with `epsilon` < 1. Energy balance may be invalid.'
            warnings.warn(msg, EquatorSolverWarning)
        lons = get_lons(n_points)

        def bfunc(phi, temperature):
            return np.vstack((func(phi, temperature)))

        def bc(ya, yb):
            return ya-yb
        y0 = np.atleast_2d(np.exp(-lons**2))
        result = solve_bvp(bfunc, bc, x=lons, y=y0)
        return result.x, result.y.T
    elif mode == 'ivp_iterate':
        if epsilon > 0.5:
            msg = 'Using method `ivp_iterate` with `epsilon` > 0.5. Energy balance may be invalid.'
            warnings.warn(msg, EquatorSolverWarning)
        lons = get_lons(n_points)
        y0 = np.atleast_1d(0.5)
        for _ in range(10):
            result = solve_ivp(func, (-np.pi, np.pi), y0=y0, t_eval=lons)
            phi = result.t
            temperature = result.y.T
            if np.abs(temperature[0]-temperature[-1])/temperature[-1] < 1e-2:
                return phi, temperature
        return phi, temperature
    elif mode == 'analytic':
        if epsilon < 10:
            msg = 'Using method `analytic` with `epsilon` < 10. Energy balance may be invalid.'
            warnings.warn(msg, EquatorSolverWarning)
        lons = get_lons(n_points)
        lon_prime = np.where(lons < -np.pi/2, lons+2*np.pi, lons)
        temperature_tild0 = np.pi**(-0.25)
        gamma = 4*temperature_tild0**3/epsilon
        temperature_tild_day = 3/4*temperature_tild0 + \
            (gamma*np.cos(lon_prime) + np.sin(lon_prime)) / (epsilon * (1+gamma**2)) + \
            np.exp(-gamma * lon_prime) / (2*epsilon *
                                          (1+gamma**2)*np.sinh(np.pi*gamma/2))
        temperature_tild_night = 3/4*temperature_tild0 + \
            np.exp(-gamma*(lon_prime-np.pi)) / \
            (2*epsilon*(1+gamma**2)*np.sinh(np.pi*gamma/2))
        day = (lon_prime >= -np.pi/2) & (lon_prime <= np.pi/2)
        temperature = deepcopy(temperature_tild_day)
        temperature[~day] = temperature_tild_night[~day]
        return lons, np.atleast_2d(temperature).T
    else:
        raise ValueError(f'Unknown mode: {mode}')


class TemperatureMap:
    """
    A map of the surface temperature.

    Parameters
    ----------
    epsilon : float
        The thermal inertia.
    t0 : astropy.units.Quantity
        The fiducial temperature.
    """
    min_temp = 40*u.K

    def __init__(
        self,
        epsilon,
        t0: u.Quantity
    ):
        n_points = 180
        if epsilon < 1:
            mode = 'ivp_reflect'
        elif epsilon < 10:
            mode = 'bvp'
        else:
            mode = 'analytic'

        phi, temperature = get_equator_curve(epsilon, n_points, mode)

        self.equator = interp1d(phi, temperature[:, 0])
        self.epsilon = epsilon
        self.t0 = t0

    def eval(
        self,
        lon: u.Quantity,
        lat: u.Quantity,
        alpha: float
    ) -> u.Quantity:
        """
        Evaluate the temperature map at a point of points.

        Parameters
        ----------
        lon : astropy.units.Quantity
            The longitude of the points to evaluate.
        lat : astropy.units.Quantity
            The latitude of the points to evaluate.
        alpha : float
            The temperature ratio between the pole and the equator

        Returns
        -------
        astropy.units.Quantity
            The surface temperature at the desired points.

        Notes
        -----
        The following math was done by Ted Johnson on 2024-01-26.
        The goal is to add a parameter in to account for latitudinal
        mixing. Here is my setup for the problem:

        .. math::
            f(x,\\alpha) ~\\text{is the temperature as a function of latitude}

        Where :math:`\\alpha` is defined such that :math:`f(\\pi/2,\\alpha) = \\alpha f(0,\\alpha)`.

        The flux at the surface is then :math:`(f(x,\\alpha))^4`. Due to the polar coordinate
        system the relative area is :math:`\\cos{x}`

        We can then say that because of the Stephan-Boltzmann law, the following
        quantity is invarient:

        .. math::
            I = \\int_{0}^{\\pi/2} (f(x,\\alpha) \\cos{x})^4 dx

        We also know that in the case of no latitudinal mixing, the temperature is:

        .. math::
            f(x,0) = \\cos(x)^{1/4}

        We can plug this in to the integral to see

        .. math::
            I = \\int_{0}^{\\pi/2} (\\cos(x)^{1/4} \\cos{x})^4 dx \\\\
                = \\int_{0}^{\\pi/2} (\\cos{x})^5 dx \\\\
                = \\frac{8}{15}

        We can then look at the other limit. Let :math:`T_0` be the temperature
        at which the equator and pole are equal. Then the temperature is

        .. math::
            f(x,1) = T_0

        And the integral is

        .. math::
            I = \\int_{0}^{\\pi/2} (T_0 \\cos{x})^4 dx = \\frac{8}{15}

        With the knowledge that :math:`\\int_{0}^{\\pi/2} (\\cos{x})^4 dx = \\frac{3\\pi}{16}`
        we find that:

        .. math::
            T_0 = \\left(\\frac{2^7}{45\\pi}\\right)^{1/4} \\\\
                = 0.9754654591261265

        We then can create a general equation for the temperature:

        .. math::
            f(x,\\alpha) = T_0 \\alpha + (1-\\alpha) \\cos(x)^{1/4}
        """
        if isinstance(lon, u.Quantity):
            lon = get_psi(lon)
        if alpha < 0 or alpha > 1:
            raise ValueError(f'alpha must be between 0 and 1, got {alpha}')
        t0 = (2**7/45/np.pi)**0.25  # see docstring
        tmap = self.t0*self.equator(lon) * \
            (t0*alpha + (1-alpha)*np.cos(lat)**0.25)
        tmap: u.Quantity = np.where(tmap > self.min_temp, tmap, self.min_temp)
        return tmap

    @classmethod
    def from_planet(
        cls,
        epsilon: float,
        star_teff: u.Quantity,
        albedo: float,
        r_star: u.Quantity,
        r_orbit: u.Quantity
    ):
        """
        Generate a `TemperatureMap` given the properties of a planet.

        Parameters
        ----------
        epsilon : float
            The thermal inertia of the planet.
        star_teff : astropy.units.Quantity
            The effective temperature of the host star.
        albedo : float
            The Bond albedo of the planet.
        r_star : astropy.units.Quantity
            The radius of the host star.
        r_orbit : astropy.units.Quantity
            The planet's orbital radius.

        """
        t0 = get_t0(star_teff, albedo, r_star, r_orbit)
        return cls(epsilon, t0)


def energy_balance_error(
    tmap: u.Quantity,
    lons: np.ndarray,
    lats: np.ndarray,
    star_teff: u.Quantity,
    albedo: float,
    r_star: u.Quantity,
    r_orbit: u.Quantity
) -> float:
    """
    Validate the energy balance.

    Parameters
    ----------
    tmap : astropy.units.Quantity
        The temperature map.
    lons : np.ndarray
        The longitude points of the map.
    lats : np.ndarray
        The latitude points of the map.
    star_teff : astropy.units.Quantity
        The effective temperature of the star.
    albedo : float
        The Bond albedo of the planet.
    r_star : astropy.units.Quantity
        The radius of the star.
    r_orbit : astropy.units.Quantity
        The orbital radius of the planet.

    Returns
    -------
    float
        The fractional error of the energy balance.
    """
    r_planet = 1*u.R_earth  # for units. This will divide out later.
    incident_flux = get_flux(star_teff, r_star, r_orbit)  # W/m2
    absorbed = (1-albedo) * incident_flux * np.pi * r_planet**2
    absorbed = absorbed.to(u.W)
    planet_area = 4*np.pi * r_planet**2
    mean_flux = (absorbed/planet_area).to(u.Unit('W m-2'))
    # pylint: disable-next=no-member
    mean_temperature = ((mean_flux/c.sigma_sb)**0.25).to(u.K)

    _, llats = np.meshgrid(lons, lats, indexing='ij')
    jacobian = np.cos(llats)
    dlon = 2*np.pi/len(lons)
    dlat = np.pi/len(lats)
    integrand = tmap**4 * jacobian
    # pylint: disable-next=no-member
    outbound_flux = (c.sigma_sb * dlon * dlat *
                     np.sum(integrand)).to(u.Unit('W m-2'))
    # pylint: disable-next=no-member
    eq_outbound_flux = np.sum(c.sigma_sb * mean_temperature**4 * jacobian * dlon*dlat)
    return (eq_outbound_flux-outbound_flux)/eq_outbound_flux


def to_pygcm(
    shape: Tuple,
    epsilon: float,
    star_teff: u.Quantity,
    r_star: u.Quantity,
    r_orbit: u.Quantity,
    lat_redistribution: float,
    p_surf: u.Quantity,
    p_stop: u.Quantity,
    wind_u: u.Quantity,
    wind_v: u.Quantity,
    albedo: u.Quantity,
    emissivity: u.Quantity,
    gamma: float,
    molecules: dict,
):
    """
    Create a `pypsg.PyGCM` object using the parameterized VSPEC GCM.

    Parameters
    ----------
    shape : tuple
        The shape of the temperature map.
    epsilon : float
        The thermal inertia of the planet.
    star_teff : astropy.units.Quantity
        The effective temperature of the host star.
    r_star : astropy.units.Quantity
        The radius of the host star.
    r_orbit : astropy.units.Quantity
        The planet's orbital radius.
    lat_redistribution : float
        The latitudinal redistribution factor.
    p_surf : astropy.units.Quantity
        The surface pressure.
    p_stop : astropy.units.Quantity
        The pressure at the top of the atmosphere.
    wind_u : astropy.units.Quantity
        The zonal wind.
    wind_v : astropy.units.Quantity
        The meridional wind.
    albedo : astropy.units.Quantity
        The Bond albedo of the planet.
    emissivity : astropy.units.Quantity
        The surface emissivity.
    gamma : float
        The surface heat transfer coefficient.
    molecules : dict
        The molecules in the atmosphere. Format is `{'H2O': 1e-2}`

    Returns
    -------
    pypsg.PyGCM
        The PyGCM object.
    """
    pressure = structure.Pressure.from_limits(p_surf, p_stop, shape)
    albedo = albedo if isinstance(albedo, u.Quantity) else albedo * u.dimensionless_unscaled
    tmap = TemperatureMap.from_planet(
        epsilon=epsilon,
        star_teff=star_teff,
        albedo=albedo.to_value(u.dimensionless_unscaled),
        r_star=r_star,
        r_orbit=r_orbit
    )
    lons = np.linspace(-np.pi, np.pi, shape[1], endpoint=False)
    lats = np.linspace(-np.pi/2, np.pi/2, shape[2], endpoint=True)
    llons, llats = np.meshgrid(lons, lats, indexing='ij')
    dat = tmap.eval(llons, llats, lat_redistribution)
    balance_err = energy_balance_error(
        dat, lons, lats, star_teff, albedo, r_star, r_orbit
    )
    if np.abs(balance_err) > 0.01:
        msg = f'Energy balance off by {100*balance_err:.1f} %. eps = {epsilon}, T0 = {lat_redistribution}'
        warnings.warn(msg, RuntimeWarning)

    tsurf = structure.SurfaceTemperature(dat)

    _wind_u = structure.Wind.constant('wind_u', wind_u, shape)
    _wind_v = structure.Wind.constant('wind_v', wind_v, shape)

    psurf = structure.SurfacePressure.from_pressure(pressure)
    _albedo = structure.Albedo.constant(albedo, shape[1:])
    _emissivity = structure.Emissivity.constant(emissivity*u.dimensionless_unscaled, shape[1:])
    temperature = structure.Temperature.from_adiabat(gamma, tsurf, pressure)
    _molecules = tuple(structure.Molecule.constant(
        name, float(abn), shape) for name, abn in molecules.items())

    return PyGCM(
        pressure,
        temperature,
        *_molecules,
        wind_u=_wind_u,
        wind_v=_wind_v,
        tsurf=tsurf,
        psurf=psurf,
        albedo=_albedo,
        emissivity=_emissivity,
        lon_start=-180,
        lat_start=-90,
        desc='A simple parameterized GCM.'
    )
